import os
import time
from pathlib import Path
from traitlets import Bool, List, Unicode, observe
from glue_jupyter.bqplot.image import BqplotImageView

from jdaviz.core.custom_traitlets import FloatHandleEmpty, IntHandleEmpty
from jdaviz.core.marks import ShadowMixin
from jdaviz.core.registries import tray_registry
from jdaviz.core.template_mixin import (PluginTemplateMixin, SelectPluginComponent,
                                        ViewerSelectMixin, DatasetMultiSelectMixin,
                                        SubsetSelectMixin, PluginTableSelectMixin,
                                        PluginPlotSelectMixin, AutoTextField,
                                        MultiselectMixin, with_spinner)
from glue.core.message import SubsetCreateMessage, SubsetDeleteMessage, SubsetUpdateMessage

from jdaviz.core.events import AddDataMessage, SnackbarMessage
from jdaviz.core.user_api import PluginUserApi

from specutils import Spectrum1D
from astropy import units as u
from astropy.nddata import CCDData

try:
    import cv2
except ImportError:
    HAS_OPENCV = False
else:
    import threading
    HAS_OPENCV = True

__all__ = ['Export']


@tray_registry('export', label="Export")
class Export(PluginTemplateMixin, ViewerSelectMixin, SubsetSelectMixin,
             DatasetMultiSelectMixin, PluginTableSelectMixin, PluginPlotSelectMixin,
             MultiselectMixin):
    """
    See the :ref:`Export Plugin Documentation <imviz-export-plot>` for more details.

    Only the following attributes and methods are available through the
    :ref:`public plugin API <plugin-apis>`:

    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.show`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.open_in_tray`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.close_in_tray`
    * ``viewer`` (:class:`~jdaviz.core.template_mixin.ViewerSelect`)
    * ``viewer_format`` (:class:`~jdaviz.core.template_mixin.SelectPluginComponent`)
    * ``dataset`` (:class:`~jdaviz.core.template_mixin.DatasetSelect`)
    * ``dataset_format`` (:class:`~jdaviz.core.template_mixin.SelectPluginComponent`)
    * ``subset`` (:class:`~jdaviz.core.template_mixin.SubsetSelect`)
    * ``subset_format`` (:class:`~jdaviz.core.template_mixin.SelectPluginComponent`)
    * ``plugin_table`` (:class:`~jdaviz.core.template_mixin.PluginTableSelect`)
    * ``plugin_table_format`` (:class:`~jdaviz.core.template_mixin.SelectPluginComponent`)
    * ``plugin_plot`` (:class:`~jdaviz.core.template_mixin.PluginPlotSelect`)
    * ``plugin_plot_format`` (:class:`~jdaviz.core.template_mixin.SelectPluginComponent`)
    * ``filename``
    * :meth:`export`
    """
    template_file = __file__, "export.vue"

    dev_multi_support = Bool(False).tag(sync=True)  # when enabling: add entries to docstring

    viewer_format_items = List().tag(sync=True)
    viewer_format_selected = Unicode().tag(sync=True)

    plugin_table_format_items = List().tag(sync=True)
    plugin_table_format_selected = Unicode().tag(sync=True)

    subset_format_items = List().tag(sync=True)
    subset_format_selected = Unicode().tag(sync=True)

    dataset_format_items = List().tag(sync=True)
    dataset_format_selected = Unicode().tag(sync=True)

    # copy of widget of the selected plugin_plot in case the parent plugin is not opened
    plugin_plot_selected_widget = Unicode().tag(sync=True)

    plugin_plot_format_items = List().tag(sync=True)
    plugin_plot_format_selected = Unicode().tag(sync=True)

    filename_value = Unicode().tag(sync=True)
    filename_default = Unicode().tag(sync=True)
    filename_auto = Bool(True).tag(sync=True)
    filename_invalid_msg = Unicode('').tag(sync=True)

    default_filepath = Unicode().tag(sync=True)

    # if selected subset is spectral or composite, display message and disable export
    subset_invalid_msg = Unicode().tag(sync=True)
    data_invalid_msg = Unicode().tag(sync=True)

    # We currently disable exporting spectrum-viewer in Cubeviz
    viewer_invalid_msg = Unicode().tag(sync=True)

    # For Cubeviz movie.
    movie_enabled = Bool(False).tag(sync=True)
    i_start = IntHandleEmpty(0).tag(sync=True)
    i_end = IntHandleEmpty(0).tag(sync=True)
    movie_fps = FloatHandleEmpty(5.0).tag(sync=True)
    movie_recording = Bool(False).tag(sync=True)
    movie_interrupt = Bool(False).tag(sync=True)

    overwrite_warn = Bool(False).tag(sync=True)

    # This is a temporary measure to allow server-installations to disable saving server-side until
    # saving client-side is supported for all exports.
    serverside_enabled = Bool(True).tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.filename = AutoTextField(self, 'filename_value',
                                      'filename_default',
                                      'filename_auto',
                                      'filename_invalid_msg')

        # description displayed under plugin title in tray
        self._plugin_description = 'Export data/plots and other outputs to a file.'

        # NOTE: if adding export support for non-plugin products, also update the language
        # in the UI as well as in _set_dataset_not_supported_msg
        self.dataset.filters = ['is_not_wcs_only', 'not_child_layer',
                                'from_plugin']

        viewer_format_options = ['png', 'svg']
        if self.config == 'cubeviz':
            if not self.app.state.settings.get('server_is_remote'):
                viewer_format_options += ['mp4']
                # still list mp4 as an option, but display a message (and raise an error) if
                # opencv is not available
                self.movie_enabled = HAS_OPENCV

        self.viewer_format = SelectPluginComponent(self,
                                                   items='viewer_format_items',
                                                   selected='viewer_format_selected',
                                                   manual_options=viewer_format_options)

        # NOTE: see self.plugin_table.selected_obj.write.list_formats() for full list of options,
        # although not all support passing overwrite
        plugin_table_format_options = ['ecsv', 'csv', 'fits']
        self.plugin_table_format = SelectPluginComponent(self,
                                                         items='plugin_table_format_items',
                                                         selected='plugin_table_format_selected',
                                                         manual_options=plugin_table_format_options)

        subset_format_options = [{'label': 'fits', 'value': 'fits', 'disabled': False},
                                 {'label': 'reg', 'value': 'reg', 'disabled': False},
                                 {'label': 'ecsv', 'value': 'ecsv', 'disabled': True}]
        self.subset_format = SelectPluginComponent(self,
                                                   items='subset_format_items',
                                                   selected='subset_format_selected',
                                                   manual_options=subset_format_options)

        dataset_format_options = ['fits']
        self.dataset_format = SelectPluginComponent(self,
                                                    items='dataset_format_items',
                                                    selected='dataset_format_selected',
                                                    manual_options=dataset_format_options)

        plugin_plot_format_options = ['png', 'svg']
        self.plugin_plot_format = SelectPluginComponent(self,
                                                        items='plugin_plot_format_items',
                                                        selected='plugin_plot_format_selected',
                                                        manual_options=plugin_plot_format_options)

        # default selection:
        self.dataset._default_mode = 'empty'
        self.subset._default_mode = 'empty'
        self.plugin_table._default_mode = 'empty'
        self.plugin_plot._default_mode = 'empty'
        self.plugin_plot.select_default()
        self.plugin_table.select_default()
        # viewer last so that the first viewer is the default and all others are empty
        self.viewer.select_default()

        if self.config == "cubeviz":
            self.session.hub.subscribe(self, AddDataMessage, handler=self._on_cubeviz_data_added)  # noqa: E501

        self.session.hub.subscribe(self, SubsetCreateMessage,
                                   handler=self._set_subset_not_supported_msg)
        self.session.hub.subscribe(self, SubsetUpdateMessage,
                                   handler=self._set_subset_not_supported_msg)
        self.session.hub.subscribe(self, SubsetDeleteMessage,
                                   handler=self._set_subset_not_supported_msg)

        if self.app.state.settings.get('server_is_remote', False):
            # when the server is remote, saving the file in python would save on the server, not
            # on the user's machine, so export support in cubeviz should be disabled
            self.serverside_enabled = False

    @property
    def user_api(self):
        # TODO: backwards compat for save_figure, save_movie,
        # i_start, i_end, movie_fps, movie_filename
        expose = ['viewer', 'viewer_format',
                  'dataset', 'dataset_format',
                  'subset', 'subset_format',
                  'plugin_table', 'plugin_table_format',
                  'plugin_plot', 'plugin_plot_format',
                  'filename', 'export']

        if self.dev_multi_support:
            expose += ['multiselect']

        return PluginUserApi(self, expose=expose)

    def _on_cubeviz_data_added(self, msg):
        # NOTE: This needs revising if we allow loading more than one cube.
        if isinstance(msg.viewer, BqplotImageView):
            if len(msg.data.shape) == 3:
                self.i_end = msg.data.shape[-1] - 1

    @observe('multiselect', 'viewer_multiselect')
    def _sync_multiselect_traitlets(self, event):
        # ViewerSelectMixin brings viewer_multiselect, but we want a single traitlet to control
        # all select inputs, so we'll keep them synced here and only expose multiselect through
        # the user API
        self.multiselect = event.get('new')
        self.viewer_multiselect = event.get('new')
        if not self.multiselect:
            # default to just a single viewer
            self._sync_singleselect({'name': 'viewer', 'new': self.viewer_selected})

    @observe('viewer_selected', 'dataset_selected', 'subset_selected',
             'plugin_table_selected', 'plugin_plot_selected')
    def _sync_singleselect(self, event):

        if not hasattr(self, 'dataset') or not hasattr(self, 'viewer'):
            # plugin not fully intialized
            return
        # if multiselect is not enabled, only allow a single selection across all select components
        if self.multiselect:
            return
        if event.get('new') in ('', []):
            return
        name = event.get('name')
        for attr in ('viewer_selected', 'dataset_selected', 'subset_selected',
                     'plugin_table_selected', 'plugin_plot_selected'):
            if name != attr:
                setattr(self, attr, '')
            if attr == 'subset_selected':
                if self.subset.selected != '':
                    self._update_subset_format_disabled()
                self._set_subset_not_supported_msg()
            if attr == 'dataset_selected':
                self._set_dataset_not_supported_msg()

    @observe('viewer_selected', 'viewer_format_selected',
             'dataset_selected', 'dataset_format_selected',
             'subset_selected', 'subset_format_selected',
             'plugin_table_selected', 'plugin_table_format_selected',
             'plugin_plot_selected', 'plugin_plot_format_selected')
    def _set_filename_default(self, *args, **kwargs):
        if not hasattr(self, 'viewer'):
            return
        if self.multiselect:
            raise NotImplementedError("batch export not yet supported")
        if self.viewer_selected:
            self.filename_default = f"{self.viewer_selected}.{self.viewer_format_selected}"  # noqa
        elif self.dataset_selected:
            self.filename_default = f"{self.dataset_selected.replace(' ', '_')}.{self.dataset_format_selected}"  # noqa
        elif self.subset_selected:
            self.filename_default = f"{self.subset_selected.replace(' ', '_')}.{self.subset_format_selected}"  # noqa
        elif self.plugin_table_selected:
            self.filename_default = f"{self.plugin_table_selected.replace(':', '').replace(' ', '_')}.{self.plugin_table_format_selected}"  # noqa
        elif self.plugin_plot_selected:
            self.filename_default = f"{self.plugin_plot_selected.replace(':', '').replace(' ', '_')}.{self.plugin_plot_format_selected}"  # noqa
        else:
            self.filename_default = ''

        # Call to initially set the default filepath
        if hasattr(self, 'viewer_format') and self.filename_default:
            self._normalize_filename(filename=self.filename_default,
                                     filetype=self.viewer_format.selected,
                                     default_path=True)

    @observe('filename_value')
    def _is_filename_changed(self, event):
        # Update the UI Filepath if relative or absolute paths are provided
        # by user via self.filename_value
        expanded_path = os.path.expanduser(self.filename_value)
        abs_path = Path(expanded_path).resolve()
        self.default_filepath = str(abs_path)

        # Clear overwrite warning when user changes filename
        self.overwrite_warn = False

    def _update_subset_format_disabled(self):
        new_items = []
        if self.subset.selected is not None:
            try:
                subset = self.app.get_subsets(self.subset.selected)
            except Exception:
                # subset invalid message will already be set,
                # no need to set valid/invalid formats.
                return
            if self.app._is_subset_spectral(subset[0]):
                good_formats = ["ecsv"]
            else:
                good_formats = ["fits", "reg"]
            for item in self.subset_format_items:
                if item["label"] in good_formats:
                    item["disabled"] = False
                else:
                    item["disabled"] = True
                    if item["label"] == self.subset_format.selected:
                        self.subset_format.selected = good_formats[0]
                new_items.append(item)
        self.subset_format_items = []
        self.subset_format_items = new_items

    @observe('subset_format_selected')
    def _disable_subset_format_combo(self, event):
        # Disable selecting a bad subset+format combination from the API
        if self.subset.selected == '' or self.subset.selected is None:
            return
        subset = self.app.get_subsets(self.subset.selected)
        bad_combo = False
        if self.app._is_subset_spectral(subset[0]):
            if event['new'] != "ecsv":
                bad_combo = True
        elif event['new'] == "ecsv":
            bad_combo = True

        if bad_combo:
            # Set back to a good value and raise error
            good_format = [format["label"] for format in self.subset_format_items if
                           format["disabled"] is False][0]
            self.subset_format.selected = good_format
            raise ValueError(f"Cannot export {self.subset.selected} in {event['new']}"
                             f" format, reverting selection to {self.subset_format.selected}")

    def _set_subset_not_supported_msg(self, msg=None):
        """
        Check if selected subset is spectral or composite, and warn and
        disable Export button until these are supported.
        """

        if self.subset.selected not in [None, '']:
            try:
                subset = self.app.get_subsets(self.subset.selected)
            except Exception as e:
                self.subset_invalid_msg = f"Export for subset not supported: {e}"
            else:
                if self.subset.selected == '':
                    self.subset_invalid_msg = ''
                elif self.app._is_subset_spectral(subset[0]):
                    self.subset_invalid_msg = ''
                elif len(subset) > 1:
                    self.subset_invalid_msg = 'Export for composite subsets not yet supported.'
                else:
                    self.subset_invalid_msg = ''
        else:  # no subset selected (can be '' instead of None if previous selection made)
            self.subset_invalid_msg = ''

    def _set_dataset_not_supported_msg(self, msg=None):
        if self.dataset.selected_obj is not None:
            if self.dataset.selected_obj.meta.get("Plugin", None) is None:
                # NOTE: should not be a valid choice due to dataset filters, but we'll include
                # another check here.
                self.data_invalid_msg = "Data export is only available for plugin generated data."
            elif not isinstance(self.dataset.selected_obj, (Spectrum1D, CCDData)):
                self.data_invalid_msg = "Export is not yet implemented for this type of data"
            elif (data_unit := self.dataset.selected_obj.unit) == u.Unit('DN/s'):
                self.data_invalid_msg = f'Export Disabled: The unit {data_unit} could not be saved in native FITS format.'  # noqa: E501
            else:
                self.data_invalid_msg = ''
        else:
            self.data_invalid_msg = ''

    def _normalize_filename(self, filename=None, filetype=None, overwrite=False, default_path=False):  # noqa: E501
        # Make sure filename is valid and file does not end up in weird places in standalone mode.
        if not filename:
            raise ValueError("Invalid filename")

        if isinstance(filename, str):
            if not filename.endswith(filetype):
                filename += f".{filetype}"
            filename = Path(filename).expanduser()

        filename = filename.resolve()
        filepath = filename.parent
        if filepath and not filepath.is_dir():
            raise ValueError(f"Invalid path={filepath}")
        elif ((not filepath or str(filepath).startswith(".")) and os.environ.get("JDAVIZ_START_DIR", "")):  # noqa: E501 # pragma: no cover
            filename = os.environ["JDAVIZ_START_DIR"] / filename

        # Set the default filepath to inform user where file will be exported to
        if default_path:
            if not filepath.is_absolute():
                abs_path = filepath.resolve()
                self.default_filepath = str(abs_path)
            # set default path for standalone
            if os.environ.get("JDAVIZ_START_DIR", ""):
                self.default_filepath = os.environ["JDAVIZ_START_DIR"]

        if filename.exists() and not overwrite and not default_path:
            self.overwrite_warn = True
        else:
            self.overwrite_warn = False

        return filename

    @with_spinner()
    def export(self, filename=None, show_dialog=None, overwrite=False,
               raise_error_for_overwrite=True):
        """
        Export selected item(s)

        Parameters
        ----------
        filename : str, optional
            If not provided, plugin value will be used.

        show_dialog : bool or `None`
            If `True`, prompts dialog to save PNG/SVG from browser.

        overwrite : bool
            If `True`, silently overwrite an existing file.

        raise_error_for_overwrite : bool
            If `True`, raise exception when ``overwrite=False`` but
            output file already exists. Otherwise, a message will be sent
            to application snackbar instead.

        """
        if self.multiselect:
            raise NotImplementedError("batch export not yet supported")

        filename = filename if filename is not None else self.filename_value

        # at this point, we can assume only a single export is selected
        if len(self.viewer.selected):
            viewer = self.viewer.selected_obj
            if len(self.viewer.selected):
                if self.viewer_invalid_msg != "":
                    raise NotImplementedError(f"Viewer cannot be exported - {self.viewer_invalid_msg}")  # noqa

            viewer = self.viewer.selected_obj

            filetype = self.viewer_format.selected
            filename = self._normalize_filename(filename, filetype, overwrite=overwrite)

            if self.overwrite_warn and not overwrite:
                if raise_error_for_overwrite:
                    raise FileExistsError(f"{filename} exists but overwrite=False")
                return

            # temporarily "clean" incompatible marks of unicode characters, etc
            restores = []
            for mark in viewer.figure.marks:
                restore = {}
                if len(getattr(mark, 'text', [])):
                    if not isinstance(mark, ShadowMixin):
                        # if it is shadowing another mark, that will automatically get updated
                        # when the other mark is restored, but we'll still ensure that the mark
                        # is clean of unicode before exporting.
                        restore['text'] = [t for t in mark.text]
                    mark.text = [t.strip() for t in mark.text]
                if len(getattr(mark, 'labels', [])):
                    restore['labels'] = mark.labels[:]
                    mark.labels = [lbl.strip() for lbl in mark.labels]
                restores.append(restore)

            if filetype == "mp4":
                self.save_movie(viewer, filename, filetype)
            else:
                self.save_figure(viewer, filename, filetype, show_dialog=show_dialog)

            # restore marks to their original state
            for restore, mark in zip(restores, viewer.figure.marks):
                for k, v in restore.items():
                    setattr(mark, k, v)

        elif len(self.plugin_plot.selected):
            plot = self.plugin_plot.selected_obj._obj
            filetype = self.plugin_plot_format.selected
            filename = self._normalize_filename(filename, filetype, overwrite=overwrite)

            if not plot._plugin.is_active:
                # force an update to the plot.  This requires the plot to have set
                # update_callback when instantiated
                plot._update()

                # create a copy of the widget shown off screen to enable rendering
                # in case one was never created in the parent plugin
                self.plugin_plot_selected_widget = f'IPY_MODEL_{plot.model_id}'

            if self.overwrite_warn and not overwrite:
                if raise_error_for_overwrite:
                    raise FileExistsError(f"{filename} exists but overwrite={overwrite}")
                return

            self.save_figure(plot, filename, filetype, show_dialog=show_dialog)

        elif len(self.plugin_table.selected):
            filetype = self.plugin_table_format.selected
            filename = self._normalize_filename(filename, filetype, overwrite=overwrite)
            if self.overwrite_warn and not overwrite:
                if raise_error_for_overwrite:
                    raise FileExistsError(f"{filename} exists but overwrite=False")
                return
            self.plugin_table.selected_obj.export_table(filename, overwrite=True)

        elif len(self.subset.selected):
            selected_subset_label = self.subset.selected
            filetype = self.subset_format.selected
            filename = self._normalize_filename(filename, filetype, overwrite=overwrite)
            if self.subset_invalid_msg != '':
                raise NotImplementedError(f'Subset can not be exported - {self.subset_invalid_msg}')
            if self.overwrite_warn and not overwrite:
                if raise_error_for_overwrite:
                    raise FileExistsError(f"{filename} exists but overwrite=False")
                return

            if self.subset_format.selected in ('fits', 'reg'):
                self.save_subset_as_region(selected_subset_label, filename)
            elif self.subset_format.selected == 'ecsv':
                self.save_subset_as_table(filename)

        elif len(self.dataset.selected):
            filetype = self.dataset_format.selected
            filename = self._normalize_filename(filename, filetype, overwrite=overwrite)
            if self.data_invalid_msg != "":
                raise NotImplementedError(f"Data can not be exported - {self.data_invalid_msg}")
            if self.overwrite_warn and not overwrite:
                if raise_error_for_overwrite:
                    raise FileExistsError(f"{filename} exists but overwrite=False")
                return
            self.dataset.selected_obj.write(Path(filename), overwrite=True)
        else:
            raise ValueError("nothing selected for export")

        return filename

    def vue_export_from_ui(self, *args, **kwargs):
        try:
            filename = self.export(show_dialog=True, raise_error_for_overwrite=False)
        except Exception as e:
            self.hub.broadcast(SnackbarMessage(
                f"Export failed with: {e}", sender=self, color="error"))
        else:
            if filename is not None:
                self.hub.broadcast(SnackbarMessage(
                    f"Exported to {filename}", sender=self, color="success"))

    def vue_overwrite_from_ui(self, *args, **kwargs):
        """Attempt to force writing the output if the user confirms the desire to overwrite."""
        try:
            filename = self.export(show_dialog=True, overwrite=True,
                                   raise_error_for_overwrite=False)
        except Exception as e:
            self.hub.broadcast(SnackbarMessage(
                f"Export with overwrite failed with: {e}", sender=self, color="error"))
        else:
            if filename is not None:
                self.hub.broadcast(SnackbarMessage(
                    f"Exported to {filename} (overwrite)", sender=self, color="success"))
            self.overwrite_warn = False

    def save_figure(self, viewer, filename=None, filetype="png", show_dialog=False):
        if filetype == "png":

            if filename is None:
                filename = self.filename_default

            def on_img_received(data):
                try:
                    with filename.open(mode='bw') as f:
                        f.write(data)
                except Exception as e:
                    self.hub.broadcast(SnackbarMessage(
                        f"{self.viewer.selected} failed to export to {str(filename)}: {e}",
                        sender=self, color="error"))
                finally:
                    self.hub.broadcast(SnackbarMessage(
                        f"{self.viewer.selected} exported to {str(filename)}",
                        sender=self, color="success"))

            if viewer.figure._upload_png_callback is not None:
                raise ValueError("previous png export is still in progress. Wait to complete before making another call to save_figure")  # noqa: E501 # pragma: no cover

            viewer.figure.get_png_data(on_img_received)

        elif filetype == "svg":
            viewer.figure.save_svg(str(filename) if filename is not None else None)

    @with_spinner('movie_recording')
    def _save_movie(self, viewer, i_start, i_end, fps, filename, rm_temp_files):
        # NOTE: All the stuff here has to be in the same thread but
        #       separate from main app thread to work.

        if not self.movie_enabled:
            if not HAS_OPENCV:
                raise ImportError("Please install opencv-python")
            raise ValueError("movie support disabled")

        slice_plg = self.app._jdaviz_helper.plugins["Slice"]._obj
        orig_slice = viewer.slice
        temp_png_files = []
        i = i_start
        video = None

        # TODO: Expose to users?
        i_step = 1  # Need n_frames check if we allow tweaking

        try:
            while i <= i_end:
                if self.movie_interrupt:
                    break

                slice_plg.vue_play_next()
                cur_pngfile = Path(f"._cubeviz_movie_frame_{i}.png")
                # TODO: skip success snackbars when exporting temp movie frames?
                self.save_figure(viewer, filename=cur_pngfile, filetype="png", show_dialog=False)
                temp_png_files.append(cur_pngfile)
                i += i_step

                # Wait for the roundtrip to the frontend to complete.
                while viewer.figure._upload_png_callback is not None:
                    time.sleep(0.05)

            if not self.movie_interrupt:
                # Grab frame size.
                frame_shape = cv2.imread(temp_png_files[0]).shape
                frame_size = (frame_shape[1], frame_shape[0])

                video = cv2.VideoWriter(filename, cv2.VideoWriter_fourcc(*'mp4v'), fps, frame_size, True)  # noqa: E501
                for cur_pngfile in temp_png_files:
                    video.write(cv2.imread(cur_pngfile))
        except Exception as err:
            self.hub.broadcast(SnackbarMessage(
                f"Error saving {filename}: {err!r}", sender=self, color="error"))
        finally:
            cv2.destroyAllWindows()
            if video:
                video.release()
            slice_plg._on_slider_updated({'new': orig_slice})

        if rm_temp_files or self.movie_interrupt:
            for cur_pngfile in temp_png_files:
                if os.path.exists(cur_pngfile):
                    os.remove(cur_pngfile)

        if self.movie_interrupt:
            if os.path.exists(filename):
                os.remove(filename)
            self.movie_interrupt = False

    def save_movie(self, viewer, filename, filetype, i_start=None, i_end=None, fps=None,
                   rm_temp_files=True):
        """Save selected slices as a movie.

        This method creates a PNG file per frame (``._cubeviz_movie_frame_<n>.png``)
        in the working directory before stitching all the frames into a movie.
        Please make sure you have sufficient memory for this operation.
        PNG files are deleted after the movie is created unless otherwise specified.
        If another PNG file with the same name already exists, it will be silently replaced.

        Parameters
        ----------
        i_start, i_end : int or `None`
            Slices to record; each slice will be a frame in the movie.
            If not given, it is obtained from plugin inputs.
            Unlike Python indexing, ``i_end`` is inclusive.
            Wrapping and reverse indexing are not supported.

        fps : float or `None`
            Frame rate in frames per second (FPS).
            If not given, it is obtained from plugin inputs.

        filename : str or `None`
            Filename for the movie to be recorded. Include path if necessary.
            If not given, it is obtained from plugin inputs.
            If another file with the same name already exists, it will be silently replaced.

        filetype : {'mp4', `None`}
            Currently only MPEG-4 is supported. This keyword is reserved for future support
            of other format(s).

        rm_temp_files : bool
            Remove temporary PNG files after movie creation. Default is `True`.

        Returns
        -------
        out_filename : str
            The absolute path to the actual output file.

        """
        if self.config != "cubeviz":
            raise NotImplementedError(f"save_movie is not available for config={self.config}")

        if not HAS_OPENCV:
            raise ImportError("Please install opencv-python to save cube as movie.")

        if filetype != "mp4":
            raise NotImplementedError(f"filetype={filetype} not supported")

        if viewer.shape is None:
            raise ValueError("Selected viewer has no display shape.")

        if fps is None:
            fps = float(self.movie_fps)
        if fps <= 0:
            raise ValueError("Invalid frame rate, must be positive non-zero value.")

        if i_start is None:
            i_start = int(self.i_start)

        if i_end is None:
            i_end = int(self.i_end)

        # No wrapping. Forward only.
        slice_plg = self.app._jdaviz_helper.plugins["Slice"]._obj
        if i_start < 0:  # pragma: no cover
            i_start = 0
        max_slice = len(slice_plg.valid_values_sorted) - 1
        if i_end > max_slice:  # pragma: no cover
            i_end = max_slice
        if i_end <= i_start:
            raise ValueError(f"No frames to write: i_start={i_start}, i_end={i_end}")

        threading.Thread(
            target=lambda: self._save_movie(viewer, i_start, i_end, fps, filename, rm_temp_files)
        ).start()

        return filename

    def save_subset_as_region(self, selected_subset_label, filename):
        """
        Save a subset to file as a Region object in the working directory.
        Currently only enabled for non-composite spatial subsets. Can be saved
        as a .fits or .reg file. If link type is currently set to 'pixel',
        then a pixel region will be saved. If link type is 'wcs', then a sky
        region will be saved. If a file with the same name already exists in the
        working directory, it will be overwriten.
        """

        # type of region saved depends on link type
        align_by = getattr(self.app, '_align_by', None)

        region = self.app.get_subsets(subset_name=selected_subset_label,
                                      include_sky_region=align_by == 'wcs')

        region = region[0][f'{"sky_" if align_by == "wcs" else ""}region']

        region.write(str(filename), overwrite=True)

    def save_subset_as_table(self, filename):
        region = self.app.get_subsets(subset_name=self.subset.selected)
        region.write(str(filename))

    def vue_interrupt_recording(self, *args):  # pragma: no cover
        self.movie_interrupt = True
        # TODO: this will need updating when batch/multiselect support is added
        self.hub.broadcast(SnackbarMessage(
            f"Movie recording interrupted by user, {self.filename_value} will be deleted.",
            sender=self, color="warning"))



import os
import warnings

import numpy as np
import astropy.units as u
from astropy.io import fits
from astropy.nddata import NDData, NDDataArray
from stdatamodels.jwst.datamodels import Level1bModel

from jdaviz.core.registries import data_parser_registry
from jdaviz.utils import (
    standardize_metadata, download_uri_to_path,
    PRIHDR_KEY, standardize_roman_metadata
)
from jdaviz.configs.imviz.plugins.parsers import _parse_image as _parse_image_imviz

try:
    from roman_datamodels import datamodels as rdd
except ImportError:
    HAS_ROMAN_DATAMODELS = False
else:
    HAS_ROMAN_DATAMODELS = True

__all__ = ['parse_data']


@data_parser_registry("ramp-data-parser")
def parse_data(app, file_obj, data_type=None, data_label=None,
               ext=None, parent=None, cache=None, local_path=None, timeout=None,
               integration=0):
    """
    Attempts to parse a data file and auto-populate available viewers in
    rampviz.

    Parameters
    ----------
    app : `~jdaviz.app.Application`
        The application-level object used to reference the viewers.
    file_obj : str
        The path to a cube-like data file.
    data_type : str, {'flux', 'mask', 'uncert'}
        The data type used to explicitly differentiate parsed data.
    data_label : str, optional
        The label to be applied to the Glue data component.
    parent : str, optional
        Data label for "parent" data to associate with the loaded data as "child".
    cache : None, bool, or str
        Cache the downloaded file if the data are retrieved by a query
        to a URL or URI.
    local_path : str, optional
        Cache remote files to this path. This is only used if data is
        requested from `astroquery.mast`.
    timeout : float, optional
        If downloading from a remote URI, set the timeout limit for
        remote requests in seconds (passed to
        `~astropy.utils.data.download_file` or
        `~astroquery.mast.Conf.timeout`).
    integration : int, optional
        JWST Level 1b products bundle multiple integrations in a time-series into the
        same ramp file. If this keyword is specified and the observations
        are JWST Level 1b products, this integration in the time series will be selected.
    """

    group_viewer_reference_name = app._jdaviz_helper._default_group_viewer_reference_name
    diff_viewer_reference_name = app._jdaviz_helper._default_diff_viewer_reference_name
    integration_viewer_reference_name = (
        app._jdaviz_helper._default_integration_viewer_reference_name
    )

    if data_type is not None and data_type.lower() not in ('flux', 'mask', 'uncert'):
        raise TypeError("Data type must be one of 'flux', 'mask', or 'uncert' "
                        f"but got '{data_type}'")

    # If the file object is an hdulist or a string, use the generic parser for
    #  fits files.
    # TODO: this currently only supports fits files. We will want to make this
    #  generic enough to work with other file types (e.g. ASDF). For now, this
    #  supports MaNGA and JWST data.
    if isinstance(file_obj, fits.hdu.hdulist.HDUList):
        _parse_hdulist(
            app, file_obj, file_name=data_label,
            group_viewer_reference_name=group_viewer_reference_name,
            diff_viewer_reference_name=diff_viewer_reference_name,
        )
    elif isinstance(file_obj, str):
        if file_obj.lower().endswith('.asdf'):
            if not HAS_ROMAN_DATAMODELS:
                raise ImportError(
                    "ASDF detected but roman-datamodels is not installed."
                )
            with rdd.open(file_obj) as pf:
                _roman_3d_to_glue_data(
                    app, pf, data_label,
                    group_viewer_reference_name=group_viewer_reference_name,
                    diff_viewer_reference_name=diff_viewer_reference_name,
                    meta=dict(pf.meta)
                )
            return

        # try parsing file_obj as a URI/URL:
        file_obj = download_uri_to_path(
            file_obj, cache=cache, local_path=local_path, timeout=timeout
        )

        file_name = os.path.basename(file_obj)

        with fits.open(file_obj) as hdulist:
            _parse_hdulist(
                app, hdulist,
                ext=ext, parent=parent,
                file_name=data_label or file_name,
                integration=integration,
                group_viewer_reference_name=group_viewer_reference_name,
                diff_viewer_reference_name=diff_viewer_reference_name,
            )

    elif isinstance(file_obj, np.ndarray) and file_obj.ndim == 3:
        # load 3D cube to group viewer
        _parse_ndarray(
            app, file_obj, data_label=data_label, data_type=data_type,
            viewer_reference_name=group_viewer_reference_name,
            meta=getattr(file_obj, 'meta')
        )

    elif isinstance(file_obj, (np.ndarray, NDData)) and file_obj.ndim in (1, 2):
        # load 1D profile(s) to integration_viewer
        _parse_ndarray(
            app, file_obj, data_label=data_label,
            viewer_reference_name=integration_viewer_reference_name,
            meta=getattr(file_obj, 'meta')
        )

    elif isinstance(file_obj, Level1bModel):
        metadata = standardize_metadata({
            key: value for key, value in file_obj.to_flat_dict(
                include_arrays=False)
            .items()
            if key.startswith('meta')
        })

        _parse_ramp_cube(
            app, file_obj.data[integration], u.DN,
            data_label or file_obj.__class__.__name__,
            group_viewer_reference_name,
            diff_viewer_reference_name,
            meta=metadata
        )

    elif HAS_ROMAN_DATAMODELS and isinstance(file_obj, rdd.DataModel):
        with rdd.open(file_obj) as pf:
            _roman_3d_to_glue_data(
                app, pf, data_label, meta=pf.meta,
                group_viewer_reference_name=group_viewer_reference_name,
                diff_viewer_reference_name=diff_viewer_reference_name,
            )

    else:
        raise NotImplementedError(f'Unsupported data format: {file_obj}')


def move_group_axis_last(x):
    # swap axes per the conventions of ramp cubes
    # (group axis comes first) and the default in
    # rampviz (group axis expected last)
    return np.transpose(x, (1, 2, 0))


def _roman_3d_to_glue_data(
    app, file_obj, data_label,
    group_viewer_reference_name=None,
    diff_viewer_reference_name=None,
    meta=None
):
    """
    Parse a Roman 3D ramp cube file (Level 1),
    usually with suffix '_uncal.asdf'.
    """
    # update viewer reference names for Roman ramp cubes:
    # app._update_viewer_reference_name()

    data = file_obj.data

    if data_label is None:
        data_label = app.return_data_label(file_obj)

    # last axis is the group axis, first two are spatial axes:
    diff_data = np.vstack([
        # begin with a group of zeros, so
        # that `diff_data.ndim == data.ndim`
        np.zeros((1, *data[0].shape)),
        np.diff(data, axis=0)
    ])

    ramp_cube_data_label = f"{data_label}[DATA]"
    ramp_diff_data_label = f"{data_label}[DIFF]"

    data_reshaped = move_group_axis_last(data)
    diff_data_reshaped = move_group_axis_last(diff_data)

    # if the ramp cube has no units, assume DN:
    if getattr(data_reshaped, 'unit', None) is None:
        assumed_unit = u.DN
        data_reshaped <<= assumed_unit
        diff_data_reshaped <<= assumed_unit

    # load these cubes into the cache:
    app._jdaviz_helper.cube_cache[ramp_cube_data_label] = NDDataArray(
        data_reshaped
    )
    app._jdaviz_helper.cube_cache[ramp_diff_data_label] = NDDataArray(
        diff_data_reshaped
    )

    if meta is not None:
        meta = standardize_roman_metadata(file_obj)

    # load these cubes into the app:
    _parse_ndarray(
        app,
        file_obj=data_reshaped,
        data_label=ramp_cube_data_label,
        viewer_reference_name=group_viewer_reference_name,
        meta=meta
    )
    _parse_ndarray(
        app,
        file_obj=diff_data_reshaped,
        data_label=ramp_diff_data_label,
        viewer_reference_name=diff_viewer_reference_name,
        meta=meta
    )

    # the default collapse function in the profile viewer is "sum",
    # but for ramp files, "median" is more useful:
    viewer = app.get_viewer('integration-viewer')
    viewer.state.function = 'median'


def _parse_hdulist(
    app, hdulist,
    ext=None, parent=None,
    file_name=None,
    integration=None,
    group_viewer_reference_name=None,
    diff_viewer_reference_name=None,
):
    if file_name is None and hasattr(hdulist, 'file_name'):
        file_name = hdulist.file_name
    else:
        file_name = file_name or "Unknown HDU object"

    hdu = hdulist[1]  # extension containing the ramp

    if hdu.header['NAXIS'] == 2:
        # this may be a calibrated Level 2 image:
        _parse_image_imviz(app, hdulist, data_label=file_name, ext=ext, parent=parent)
        new_viewer_name = 'level-2'

        # create a level-2 viewer if none exists:
        if new_viewer_name not in app.get_viewer_reference_names():
            app._jdaviz_helper.create_image_viewer(viewer_name=new_viewer_name)

        # add the SCI extension to the level-2 viewer:
        if not ext:
            idx = 1
        elif ext and ('SCI' in ext or ext == '*'):
            idx = len(ext) - ext.index('SCI')

        app.add_data_to_viewer(new_viewer_name, app.data_collection[-idx].label)
        return

    elif hdu.header['NAXIS'] != 4:
        raise ValueError(f"Expected a ramp with NAXIS=4 (with dimensions: "
                         f"integrations, groups, x, y), but got "
                         f"NAXIS={hdu.header['NAXIS']}.")

    if 'BUNIT' in hdu.header:
        try:
            flux_unit = u.Unit(hdu.header['BUNIT'])
        except Exception:
            warnings.warn("Invalid BUNIT, using DN as data unit", UserWarning)
            flux_unit = u.DN
    else:
        warnings.warn("Invalid BUNIT, using DN as data unit", UserWarning)
        flux_unit = u.DN

    # index the ramp array by the integration to load. returns all groups and pixels.
    # cast from uint16 to integers:
    ramp_cube = hdu.data[integration].astype(int)

    metadata = standardize_metadata(hdu.header)
    if hdu.name != 'PRIMARY' and 'PRIMARY' in hdulist:
        metadata[PRIHDR_KEY] = standardize_metadata(hdulist['PRIMARY'].header)

    _parse_ramp_cube(
        app, ramp_cube, flux_unit, file_name,
        group_viewer_reference_name,
        diff_viewer_reference_name,
        meta=metadata
    )


def _parse_ramp_cube(app, ramp_cube_data, flux_unit, file_name,
                     group_viewer_reference_name, diff_viewer_reference_name,
                     meta=None):
    # last axis is the group axis, first two are spatial axes:
    diff_data = np.vstack([
        # begin with a group of zeros, so
        # that `diff_data.ndim == data.ndim`
        np.zeros((1, *ramp_cube_data[0].shape)),
        np.diff(ramp_cube_data, axis=0)
    ])

    ramp_cube = NDDataArray(move_group_axis_last(ramp_cube_data), unit=flux_unit, meta=meta)
    diff_cube = NDDataArray(move_group_axis_last(diff_data), unit=flux_unit, meta=meta)

    group_data_label = app.return_data_label(file_name, ext="DATA")
    diff_data_label = app.return_data_label(file_name, ext="DIFF")

    for data_entry, data_label, viewer_ref in zip(
            (ramp_cube, diff_cube),
            (group_data_label, diff_data_label),
            (group_viewer_reference_name, diff_viewer_reference_name)
    ):
        app.add_data(data_entry, data_label)
        app.add_data_to_viewer(viewer_ref, data_label)

        # load these cubes into the cache:
        app._jdaviz_helper.cube_cache[data_label] = data_entry

        # set as reference data in this viewer
        viewer = app.get_viewer(viewer_ref)
        viewer.state.reference_data = app.data_collection[data_label]

    app._jdaviz_helper._loaded_flux_cube = app.data_collection[group_data_label]


def _parse_ndarray(
    app, file_obj, data_label=None,
    viewer_reference_name=None,
    meta=None
):
    if data_label is None:
        data_label = app.return_data_label(file_obj)

    # Cannot change axis to ensure roundtripping within Rampviz.
    # Axes must already be (x, y, z) at this point.

    if isinstance(file_obj, NDData):
        ndd = file_obj
    else:
        ndd = NDDataArray(data=file_obj, meta=meta)
    app.add_data(ndd, data_label)

    app.add_data_to_viewer(viewer_reference_name, data_label)
    app._jdaviz_helper._loaded_flux_cube = app.data_collection[data_label]



import os

import astropy.units as u
import numpy as np
from numpy.testing import assert_allclose
import pytest
from specutils import Spectrum1D

from jdaviz.core.custom_units_and_equivs import PIX2, SPEC_PHOTON_FLUX_DENSITY_UNITS
from jdaviz.core.marks import MarkersMark
from jdaviz.configs.imviz.tests.utils import BaseImviz_WCS_NoWCS


def _get_markers_from_viewer(viewer):
    return [m for m in viewer.figure.marks if isinstance(m, MarkersMark)][0]


def _assert_dict_allclose(dict1, dict2):
    assert dict1.keys() == dict2.keys()
    for k, v in dict1.items():
        if isinstance(v, float):
            assert_allclose(v, dict2.get(k))
        elif isinstance(v, (tuple, list)):
            assert_allclose(np.asarray(v), np.asarray(dict2.get(k)))
        else:
            assert v == dict2.get(k)


def test_markers_cubeviz(tmp_path, cubeviz_helper, spectrum1d_cube):
    cubeviz_helper.load_data(spectrum1d_cube, "test")
    fv = cubeviz_helper.app.get_viewer('flux-viewer')
    sv = cubeviz_helper.app.get_viewer('spectrum-viewer')
    sb_unit = 'Jy / pix2'  # cubes loaded in Jy have sb unit of Jy / pix2
    flux_unit = 'Jy'

    label_mouseover = cubeviz_helper.app.session.application._tools['g-coords-info']

    mp = cubeviz_helper.plugins['Markers']
    mp.keep_active = True
    exp = cubeviz_helper.plugins['Export']

    # no marks yet, so table does not yet appear in export plugin
    assert "Markers: table" not in exp.plugin_table.choices

    # test event in flux viewer
    label_mouseover._viewer_mouse_event(fv,
                                        {'event': 'mousemove',
                                         'domain': {'x': 0, 'y': 0}})

    assert label_mouseover.as_text() == (f'Pixel x=00.0 y=00.0 Value +8.00000e+00 {sb_unit}',
                                         'World 13h39m59.9731s +27d00m00.3600s (ICRS)',
                                         '204.9998877673 27.0001000000 (deg)')

    _assert_dict_allclose(label_mouseover.as_dict(), {'axes_x': 0,
                                                      'axes_x:unit': 'pix',
                                                      'axes_y': 0,
                                                      'axes_y:unit': 'pix',
                                                      'data_label': 'test[FLUX]',
                                                      'slice': 1.0,
                                                      'spectral_axis': 4.62360027696835e-07,
                                                      'spectral_axis:unit': 'm',
                                                      'world_ra': 204.99988776727642,
                                                      'world_dec': 27.000099999955538,
                                                      'world:unreliable': False,
                                                      'pixel_x': 0,
                                                      'pixel_y': 0,
                                                      'pixel:unreliable': False,
                                                      'value': 8.0,
                                                      'value:unit': sb_unit,
                                                      'value:unreliable': False})

    mp._obj._on_viewer_key_event(fv, {'event': 'keydown',
                                      'key': 'm'})
    assert len(mp.export_table()) == 1
    assert len(_get_markers_from_viewer(fv).x) == 1

    # test event in spectrum viewer (with auto layer detection)
    # x = [4.62280007e-07, 4.62360028e-07]
    # y = [28, 92] Jy / pix2
    label_mouseover._viewer_mouse_event(sv,
                                        {'event': 'mousemove',
                                         'domain': {'x': 4.623e-7, 'y': 0}})

    assert label_mouseover.as_text() == (f'Cursor 4.62300e-07, 0.00000e+00 Value +8.00000e+00 {sb_unit}',  # noqa
                                         'Wave 4.62280e-07 m (0 pix)',
                                         f'Flux 2.80000e+01 {flux_unit}')
    assert label_mouseover.as_dict() == {'data_label': 'Spectrum (sum)',
                                         'axes_x': 4.622800069238093e-07,
                                         'axes_x:unit': 'm',
                                         'slice': 0.0,
                                         'spectral_axis': 4.622800069238093e-07,
                                         'spectral_axis:unit': 'm',
                                         'axes_y': 28.0,
                                         'axes_y:unit': flux_unit,
                                         'value': 28.0,
                                         'value:unit': flux_unit}

    mp._obj._on_viewer_key_event(sv, {'event': 'keydown',
                                      'key': 'm'})
    assert len(mp.export_table()) == 2
    assert len(_get_markers_from_viewer(fv).x) == 1
    assert len(_get_markers_from_viewer(sv).x) == 1

    # test event in spectrum viewer (with cursor only)
    label_mouseover.dataset_selected = 'none'
    label_mouseover._viewer_mouse_event(sv,
                                        {'event': 'mousemove',
                                         'domain': {'x': 4.623e-7, 'y': 0}})

    assert label_mouseover.as_text() == (f'Cursor 4.62300e-07, 0.00000e+00 Value +8.00000e+00 {sb_unit}',  # noqa
                                         '', '')
    assert label_mouseover.as_dict() == {'axes_x': 4.623e-07,
                                         'axes_x:unit': 'm',
                                         'axes_y': 0,
                                         'axes_y:unit': flux_unit,
                                         'data_label': '',
                                         'spectral_axis': 4.623e-07,
                                         'spectral_axis:unit': 'm',
                                         'value': 0,
                                         'value:unit': flux_unit}

    mp._obj._on_viewer_key_event(sv, {'event': 'keydown',
                                      'key': 'm'})

    # test that markers update on unit conversion
    uc = cubeviz_helper.plugins['Unit Conversion']
    uc.flux_unit.selected = 'MJy'

    # find the index of the marker's x-coordinate in the spectral axis of the original input data
    spec = cubeviz_helper.get_data('Spectrum (sum)', use_display_units=True)
    marker_index = np.where(spec.spectral_axis.value == mp._obj.marks['cubeviz-2'].x)
    # use the index to find the associated flux value of the original input
    flux_value = spec.flux[marker_index].value

    # compare the marker's (y) flux value with the flux value of the original input data
    assert_allclose(mp._obj.marks['cubeviz-2'].y[0], flux_value)

    # now check if marks update with a unit that requires spectral density equivalency
    uc.flux_unit.selected = 'erg / (Angstrom s cm2)'

    spec = cubeviz_helper.get_data('Spectrum (sum)', use_display_units=True)
    flux_value = spec.flux[marker_index].value

    assert_allclose(mp._obj.marks['cubeviz-2'].y[0], flux_value)

    assert len(mp.export_table()) == 3
    assert len(_get_markers_from_viewer(fv).x) == 1
    assert len(_get_markers_from_viewer(sv).x) == 2

    # markers hide when plugin closed and keep_active = False
    mp.keep_active = False
    assert _get_markers_from_viewer(fv).visible is False
    assert _get_markers_from_viewer(sv).visible is False

    # markers re-appear when plugin re-opened
    mp._obj.plugin_opened = True
    assert _get_markers_from_viewer(fv).visible is True
    assert _get_markers_from_viewer(sv).visible is True
    assert len(_get_markers_from_viewer(fv).x) == 1
    assert len(_get_markers_from_viewer(sv).x) == 2

    # appears as option in export plugin and exports successfully
    assert "Markers: table" in exp.plugin_table.choices
    filename = str(tmp_path / "cubeviz_export.ecsv")
    exp.filename.auto = False
    exp.filename.value = filename
    exp.plugin_table = "Markers: table"
    exp.export()
    assert os.path.isfile(filename)

    # Also exports to CSV
    filename_2 = str(tmp_path / "cubeviz_export.csv")
    exp.plugin_table_format.selected = 'csv'
    exp.filename.value = filename_2
    exp.export()
    assert os.path.isfile(filename_2)

    # clearing table clears markers
    mp.clear_table()
    assert mp.export_table() is None
    assert len(_get_markers_from_viewer(fv).x) == 0
    assert len(_get_markers_from_viewer(sv).x) == 0


@pytest.mark.parametrize("flux_unit", [u.Unit(x) for x in SPEC_PHOTON_FLUX_DENSITY_UNITS])
@pytest.mark.parametrize("angle_unit", [u.sr, PIX2])
@pytest.mark.parametrize("new_flux_unit", [u.Unit(x) for x in SPEC_PHOTON_FLUX_DENSITY_UNITS])
def test_markers_cubeviz_flux_unit_conversion(cubeviz_helper,
                                              spectrum1d_cube_custom_fluxunit,
                                              flux_unit, angle_unit, new_flux_unit):
    """
    Test the markers plugin with all possible unit conversions for
    cubes in spectral/photon surface brightness units (e.g. Jy/sr, Jy/pix2).

    The markers plugin should respect the choice of flux and angle
    unit selected in the Unit Conversion plugin, and inputs and results should
    be converted based on selection. All conversions between units in the
    flux dropdown menu in the unit conversion plugin should be supported.
    """

    if new_flux_unit == flux_unit:  # skip 'converting' to same unit
        return

    new_flux_unit_str = new_flux_unit.to_string()

    # load cube with specified unit
    cube = spectrum1d_cube_custom_fluxunit(fluxunit=flux_unit / angle_unit,
                                           shape=(5, 5, 4),
                                           with_uncerts=True)
    cubeviz_helper.load_data(cube, data_label="test")

    # get plugins
    mp = cubeviz_helper.plugins['Markers']
    uc = cubeviz_helper.plugins['Unit Conversion']._obj

    mp.keep_active = True

    fv = cubeviz_helper.app.get_viewer('flux-viewer')
    label_mouseover = cubeviz_helper.app.session.application._tools['g-coords-info']
    label_mouseover._viewer_mouse_event(fv,
                                        {'event': 'mousemove',
                                         'domain': {'x': 0, 'y': 0}})
    mp._obj._on_viewer_key_event(fv, {'event': 'keydown',
                                      'key': 'm'})

    # set to new unit
    uc.flux_unit.selected = new_flux_unit_str
    new_cube_unit_str = (new_flux_unit / angle_unit).to_string()

    # add a new marker at the same location
    label_mouseover._viewer_mouse_event(fv,
                                        {'event': 'mousemove',
                                         'domain': {'x': 0,
                                                    'y': 0}})
    # mouseover should have changed to new unit
    assert label_mouseover.as_dict()['value:unit'] == new_cube_unit_str

    mp._obj._on_viewer_key_event(fv, {'event': 'keydown',
                                      'key': 'm'})

    # make sure last marker added to table reflects new unit selection
    last_row = mp.export_table()[-1]
    assert last_row['value:unit'] == new_cube_unit_str


def test_markers_specviz2d_unit_conversion(specviz2d_helper, spectrum2d):
    data = np.zeros((5, 10))
    data[3] = np.arange(10)
    spectrum2d = Spectrum1D(flux=data*u.MJy, spectral_axis=data[3]*u.AA)
    specviz2d_helper.load_data(spectrum2d)

    uc = specviz2d_helper.plugins['Unit Conversion']
    uc.open_in_tray()
    mp = specviz2d_helper.plugins['Markers']
    mp.keep_active = True

    label_mouseover = specviz2d_helper.app.session.application._tools["g-coords-info"]
    viewer2d = specviz2d_helper.app.get_viewer("spectrum-2d-viewer")
    label_mouseover._viewer_mouse_event(viewer2d, {"event": "mousemove",
                                                   "domain": {"x": 6, "y": 3}})
    assert label_mouseover.as_text() == ('Pixel x=06.0 y=03.0 Value +6.00000e+00 MJy',
                                         'Wave 6.00000e+00 Angstrom',
                                         '')
    mp._obj._on_viewer_key_event(viewer2d, {'event': 'keydown',
                                            'key': 'm'})

    # make sure last marker added to table reflects new unit selection
    last_row = mp.export_table()[-1]
    assert last_row['value:unit'] == uc.flux_unit
    assert last_row['spectral_axis:unit'] == uc.spectral_unit

    # ensure marks work with flux conversion where spectral axis is required and
    # spectral axis conversion
    uc.flux_unit = 'erg / (Angstrom s cm2)'
    uc.spectral_unit = 'Ry'
    label_mouseover._viewer_mouse_event(viewer2d, {"event": "mousemove",
                                                   "domain": {"x": 4, "y": 3}})
    assert label_mouseover.as_text() == ('Pixel x=04.0 y=03.0 Value +7.49481e+00 erg / (Angstrom s cm2)',  # noqa
                                         'Wave 2.27817e+02 Ry',
                                         '')
    mp._obj._on_viewer_key_event(viewer2d, {'event': 'keydown',
                                            'key': 'm'})

    # make sure last marker added to table reflects new unit selection
    last_row = mp.export_table()[-1]
    assert last_row['value:unit'] == uc.flux_unit
    assert last_row['spectral_axis:unit'] == uc.spectral_unit

    second_marker_flux_unit = uc.flux_unit.selected
    second_marker_spectral_unit = uc.spectral_unit.selected

    # test edge case two non-native spectral axis required conversions
    uc.flux_unit = 'ph / (Angstrom s cm2)'
    uc.spectral_unit = 'eV'

    # make sure table flux and spectral unit doesn't update
    assert last_row['value:unit'] == second_marker_flux_unit
    assert last_row['spectral_axis:unit'] == second_marker_spectral_unit


class TestImvizMultiLayer(BaseImviz_WCS_NoWCS):
    def test_markers_layer_cycle(self):
        label_mouseover = self.imviz.app.session.application._tools['g-coords-info']

        mp = self.imviz.plugins['Markers']
        mp._obj.plugin_opened = True

        # cycle through dataset options (used for both coords info and markers)
        assert label_mouseover.dataset.choices == ['auto', 'none',
                                                   'has_wcs[SCI,1]',
                                                   'no_wcs[SCI,1]']
        assert label_mouseover.dataset.selected == 'auto'

        # auto (top, no WCS) layer
        label_mouseover._viewer_mouse_event(self.viewer,
                                            {'event': 'mousemove',
                                             'domain': {'x': 5, 'y': 5}})

        assert label_mouseover.as_text() == ('Pixel x=05.0 y=05.0 Value +5.50000e+01', '', '')
        assert label_mouseover.as_dict() == {'axes_x': 5,
                                             'axes_x:unit': 'pix',
                                             'axes_y': 5,
                                             'axes_y:unit': 'pix',
                                             'data_label': 'no_wcs[SCI,1]',
                                             'pixel_x': 5.0,
                                             'pixel_y': 5.0,
                                             'pixel:unreliable': False,
                                             'value': 55.0,
                                             'value:unit': '',
                                             'value:unreliable': False}

        mp._obj._on_viewer_key_event(self.viewer, {'event': 'keydown',
                                                   'key': 'm'})
        assert len(_get_markers_from_viewer(self.viewer).x) == 1

        # no layer (cursor position only)
        label_mouseover.dataset.select_next()
        assert label_mouseover.dataset.selected == 'none'

        label_mouseover._viewer_mouse_event(self.viewer,
                                            {'event': 'mousemove',
                                             'domain': {'x': 5, 'y': 5}})

        assert label_mouseover.as_text() == ('Pixel x=05.0 y=05.0', '', '')
        assert label_mouseover.as_dict() == {'axes_x': 5,
                                             'axes_x:unit': 'pix',
                                             'axes_y': 5,
                                             'axes_y:unit': 'pix',
                                             'data_label': '',
                                             'pixel_x': 5.0,
                                             'pixel_y': 5.0,
                                             'pixel:unreliable': False}

        mp._obj._on_viewer_key_event(self.viewer, {'event': 'keydown',
                                                   'key': 'm'})
        assert len(_get_markers_from_viewer(self.viewer).x) == 2

        # non-default layer (with WCS)
        label_mouseover.dataset.selected = 'has_wcs[SCI,1]'
        label_mouseover._viewer_mouse_event(self.viewer,
                                            {'event': 'mousemove',
                                             'domain': {'x': 5, 'y': 5}})

        assert label_mouseover.as_text() == ('Pixel x=05.0 y=05.0 Value +5.50000e+01',
                                             'World 22h30m04.5107s -20d49m54.9990s (ICRS)',
                                             '337.5187947654 -20.8319441647 (deg)')
        _assert_dict_allclose(label_mouseover.as_dict(), {'axes_x': 5,
                                                          'axes_x:unit': 'pix',
                                                          'axes_y': 5,
                                                          'axes_y:unit': 'pix',
                                                          'data_label': 'has_wcs[SCI,1]',
                                                          'world_ra': 337.5187947653852,
                                                          'world_dec': -20.831944164705973,
                                                          'world:unreliable': False,
                                                          'pixel_x': 5.0,
                                                          'pixel_y': 5.0,
                                                          'pixel:unreliable': False,
                                                          'value': 55.0,
                                                          'value:unit': '',
                                                          'value:unreliable': False})

        mp._obj._on_viewer_key_event(self.viewer, {'event': 'keydown',
                                                   'key': 'm'})
        assert len(_get_markers_from_viewer(self.viewer).x) == 3

    def test_markers_custom_viewer(self):
        label_mouseover = self.imviz.app.session.application._tools['g-coords-info']

        mp = self.imviz.plugins['Markers']
        mp._obj.plugin_opened = True

        nv = self.imviz.create_image_viewer()
        self.imviz.app.add_data_to_viewer('imviz-1', 'has_wcs[SCI,1]')

        assert label_mouseover.dataset.choices == ['auto', 'none',
                                                   'has_wcs[SCI,1]',
                                                   'no_wcs[SCI,1]']
        assert label_mouseover.dataset.selected == 'auto'

        # top-layer in default viewer is no_wcs[SCI,1]
        label_mouseover._viewer_mouse_event(self.viewer,
                                            {'event': 'mousemove',
                                             'domain': {'x': 5, 'y': 5}})
        assert label_mouseover.as_dict()['data_label'] == 'no_wcs[SCI,1]'

        # top-layer in new viewer is has_wcs[SCI,1]
        label_mouseover._viewer_mouse_event(nv,
                                            {'event': 'mousemove',
                                             'domain': {'x': 5, 'y': 5}})
        assert label_mouseover.as_dict()['data_label'] == 'has_wcs[SCI,1]'

        # choosing a dataset that is in one viewer but not the other
        label_mouseover.dataset.selected = 'no_wcs[SCI,1]'
        label_mouseover._viewer_mouse_event(self.viewer,
                                            {'event': 'mousemove',
                                             'domain': {'x': 5, 'y': 5}})
        assert label_mouseover.as_dict()['data_label'] == 'no_wcs[SCI,1]'

        label_mouseover._viewer_mouse_event(nv,
                                            {'event': 'mousemove',
                                             'domain': {'x': 5, 'y': 5}})
        assert label_mouseover.as_dict()['data_label'] == ''



'''
This file tests four tasks of the catalogs.py plugin in Imviz.
1. When no image/data is provided, the search does not crash
- Additionally there are no results available as the search did not occur
2. When an image/data is provided, specifically with no results being found,
results are available and are set to 0
3. When an image/data is provided, specifically with results being found,
results are available and are set to > 0
4. When the markers are cleared, there are no results available

The data used for testing are provided in the comments below.

Currently, these tests are driven by an SDSS search.
As more catalogs are added to the plugin, each test will need to be
specific based on the respective catalog.

Additionally, tests that check for results > 0 are correct as the data was
found from an SDSS server (and therefore should have results).
However, the number of results is determined on the belief that the search
is correct and the removal of the points outside the zoom limits is correctly done.
This may need to instead be tested for in the future.
'''

import numpy as np
from numpy.testing import assert_allclose
import pytest

from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.nddata import NDData
from astropy.table import Table, QTable


@pytest.mark.remote_data
class TestCatalogs:
    # testing that the plugin search does not crash when no data/image is provided
    def test_plugin_no_image(self, imviz_helper):
        catalogs_plugin = imviz_helper.plugins["Catalog Search"]._obj
        catalogs_plugin.plugin_opened = True
        # running the search without any data loaded into Imviz
        catalogs_plugin.vue_do_search()

        assert not catalogs_plugin.results_available

    # testing that variables update correctly when the image/data provided does not have results
    def test_plugin_image_no_result(self, imviz_helper, image_2d_wcs):
        arr = np.ones((10, 10))
        ndd = NDData(arr, wcs=image_2d_wcs)

        imviz_helper.load_data(ndd, data_label='no_results_data')

        catalogs_plugin = imviz_helper.plugins["Catalog Search"]._obj
        catalogs_plugin.plugin_opened = True
        catalogs_plugin.vue_do_search()

        # number of results should be 0
        assert catalogs_plugin.results_available
        assert catalogs_plugin.number_of_results == 0

    # testing that every variable updates correctly when the image/data provided does have results
    # data used: information based on this image -
    # https://dr12.sdss.org/fields/runCamcolField?field=76&camcol=5&run=7674
    # the z-band FITS image was downloaded and used
    # NOTE: We mark "slow" so it only runs on the dev job that is allowed to fail.
    @pytest.mark.slow
    def test_plugin_image_with_result(self, imviz_helper, tmp_path):
        arr = np.ones((1489, 2048))

        # Since we are not really displaying, need this to test zoom.
        viewer = imviz_helper.default_viewer._obj
        viewer.shape = (100, 100)
        viewer.state._set_axes_aspect_ratio(1)

        # header is based on the data provided above
        hdu1 = fits.ImageHDU(arr, name='SCI')
        hdu1.header.update({'CTYPE1': 'RA---TAN',
                            'CUNIT1': 'deg',
                            'CD1_1': -7.80378407867e-05,
                            'CD1_2': 7.74904339463e-05,
                            'CRPIX1': 1025.0,
                            'CRVAL1': 6.62750450757,
                            'NAXIS1': 2048,
                            'CTYPE2': 'DEC--TAN',
                            'CUNIT2': 'deg',
                            'CD2_1': 7.74973322238e-05,
                            'CD2_2': 7.80788034973e-05,
                            'CRPIX2': 745.0,
                            'CRVAL2': 1.54470013629,
                            'NAXIS2': 1489})
        imviz_helper.load_data(hdu1, data_label='has_wcs')

        catalogs_plugin = imviz_helper.plugins["Catalog Search"]._obj
        catalogs_plugin.plugin_opened = True

        # test SDSS catalog
        catalogs_plugin.catalog.selected = 'SDSS'

        # testing that SDSS catalog respects the maximum sources set
        catalogs_plugin.max_sources = 100
        catalogs_plugin.search(error_on_fail=True)

        assert catalogs_plugin.results_available
        assert catalogs_plugin.number_of_results == catalogs_plugin.max_sources

        # reset max_sources to it's default value
        catalogs_plugin.max_sources = 1000
        # This basically calls the following under the hood:
        #   skycoord_center = SkyCoord(6.62754354, 1.54466139, unit="deg")
        #   zoom_radius = r_max = 3 * u.arcmin
        #   query_region_result = SDSS.query_region(skycoord_center, radius=zoom_radius, ...)

        # Zoom in so we have to filter sources outside the viewer bounds
        viewer = imviz_helper.app.get_viewer('imviz-0')
        viewer.state.x_min = 800
        viewer.state.x_max = 1200
        viewer.state.y_min = 800
        viewer.state.y_max = 1200

        catalogs_plugin.search(error_on_fail=True)

        assert catalogs_plugin.catalog.selected == 'SDSS'

        assert catalogs_plugin.results_available
        assert catalogs_plugin.number_of_results == 136
        prev_results = catalogs_plugin.number_of_results
        last_row = catalogs_plugin.table.items[-1]
        last_ra = float(last_row['Right Ascension (degrees)'])
        coords = viewer.state.reference_data.coords
        table_calc_ra = coords.pixel_to_world(float(last_row['x_coord']), float(last_row['y_coord'])).ra.value  # noqa
        assert_allclose(last_ra, table_calc_ra, rtol=0.00001)
        assert_allclose(last_ra, imviz_helper.app._catalog_source_table[-1]['ra'], atol=0.0001)  # noqa

        # testing that every variable updates accordingly when markers are cleared
        catalogs_plugin.clear_table()

        # Default zoom for the rest of the tests
        viewer.state.x_min = -0.5
        viewer.state.x_max = 2047.5
        viewer.state.y_min = -0.5
        viewer.state.y_max = 1488.5

        assert not catalogs_plugin.results_available

        # test loading from file
        table = imviz_helper.app._catalog_source_table
        qtable = QTable({'sky_centroid': SkyCoord(table['ra'], table['dec'], unit='deg'),
                         'label': table['objid']})
        tmp_file = tmp_path / 'test.ecsv'
        qtable.write(tmp_file, overwrite=True)

        # reset max_sources to it's default value
        catalogs_plugin.max_sources = 1000

        catalogs_plugin.from_file = str(tmp_file)
        # setting filename from API will automatically set catalog to 'From File...'
        assert catalogs_plugin.catalog.selected == 'From File...'
        catalogs_plugin.search(error_on_fail=True)
        assert catalogs_plugin.results_available
        assert catalogs_plugin.number_of_results == prev_results

        catalogs_plugin.select_rows(slice(0, 2))
        assert len(catalogs_plugin.table.selected_rows) == 2

        # test Gaia catalog
        catalogs_plugin.catalog.selected = 'Gaia'

        assert catalogs_plugin.catalog.selected == 'Gaia'

        # astroquery.gaia query has the Gaia.ROW_LIMIT parameter that limits the number of rows
        # returned. Test to verify that this query functionality is maintained by the package.
        # Note: astroquery.sdss does not have this parameter.
        catalogs_plugin.max_sources = 10
        with pytest.warns(ResourceWarning):
            catalogs_plugin.search(error_on_fail=True)

        assert catalogs_plugin.results_available
        assert catalogs_plugin.number_of_results == catalogs_plugin.max_sources

        assert imviz_helper.viewers['imviz-0']._obj.state.x_min == 279.0
        assert imviz_helper.viewers['imviz-0']._obj.state.x_max == 1768.0
        assert imviz_helper.viewers['imviz-0']._obj.state.y_min == -0.5
        assert imviz_helper.viewers['imviz-0']._obj.state.y_max == 1488.5

        # Re-populate the table with a new search
        with pytest.warns(ResourceWarning):
            catalogs_plugin.search(error_on_fail=True)
        # Ensure at least one row is selected before zooming
        catalogs_plugin.table.selected_rows = [catalogs_plugin.table.items[0]]
        assert len(catalogs_plugin.table.selected_rows) > 0

        # set 'padding' to reproduce original hard-coded 50 pixel window
        # so test results don't change
        catalogs_plugin.zoom_to_selected(padding=50 / 2048)

        assert_allclose(
            imviz_helper.viewers['imviz-0']._obj.state.x_min, 1046.377555, atol=0.1)
        assert_allclose(
            imviz_helper.viewers['imviz-0']._obj.state.x_max, 1098.748805, atol=0.1)
        assert_allclose(
            imviz_helper.viewers['imviz-0']._obj.state.y_min, 699.110485, atol=0.1)
        assert_allclose(
            imviz_helper.viewers['imviz-0']._obj.state.y_max, 751.481735, atol=0.1)


def test_from_file_parsing(imviz_helper, tmp_path):
    catalogs_plugin = imviz_helper.plugins["Catalog Search"]._obj

    # _on_file_path_changed is fired when changing the selection in the file dialog
    catalogs_plugin.catalog._on_file_path_changed({'new': './invalid_path'})
    assert catalogs_plugin.from_file_message == 'File path does not exist'

    # observe('from_file') is fired when setting from_file from the API or via import_file
    # (or after clicking select in the file dialog)
    with pytest.raises(ValueError, match='./invalid_path is not a valid file path'):
        catalogs_plugin.import_catalog('./invalid_path')

    # setting to a blank string from the API resets the catalog selection to the
    # default/first entry
    catalogs_plugin.from_file = ''
    assert catalogs_plugin.catalog.selected == catalogs_plugin.catalog.choices[0]

    not_table_file = tmp_path / 'not_table.tst'
    not_table_file.touch()
    catalogs_plugin.catalog._on_file_path_changed({'new': not_table_file})
    assert catalogs_plugin.from_file_message == 'Could not parse file with astropy.table.QTable.read'  # noqa

    qtable = QTable({'not_sky_centroid': [1, 2, 3]})
    not_valid_table = tmp_path / 'not_valid_table.ecsv'
    qtable.write(not_valid_table, overwrite=True)
    catalogs_plugin.catalog._on_file_path_changed({'new': not_valid_table})
    assert catalogs_plugin.from_file_message == 'Table does not contain required sky_centroid column'  # noqa


def test_offline_ecsv_catalog(imviz_helper, image_2d_wcs):
    # Since we are not really displaying, need this to test zoom.
    viewer = imviz_helper.default_viewer._obj
    viewer.shape = (100, 100)
    viewer.state._set_axes_aspect_ratio(1)

    sky = SkyCoord(ra=[337.5202807, 337.51909197, 337.51760596],
                   dec=[-20.83305528, -20.83222194, -20.83083304], unit='deg')
    tbl = QTable({'sky_centroid': sky})  # Table has no "Label" column
    n_entries = len(tbl)

    ndd = NDData(np.ones((10, 10)), wcs=image_2d_wcs)
    imviz_helper.load_data(ndd, data_label='data_with_wcs')
    assert len(imviz_helper.app.data_collection) == 1

    catalogs_plugin = imviz_helper.plugins['Catalog Search']
    catalogs_plugin.import_catalog(tbl)
    out_tbl = catalogs_plugin._obj.search(error_on_fail=True)
    assert len(out_tbl) == n_entries
    assert catalogs_plugin._obj.number_of_results == n_entries
    # Assert that Object ID is set to index + 1 when the label column is absent
    for idx, item in enumerate(catalogs_plugin._obj.table.items):
        assert item['Object ID'] == str(idx + 1)
    assert len(imviz_helper.app.data_collection) == 2  # image + markers

    catalogs_plugin._obj.table.selected_rows = [catalogs_plugin._obj.table.items[0]]
    assert len(catalogs_plugin._obj.table.selected_rows) == 1

    # test to ensure sources searched for respect the maximum sources traitlet
    catalogs_plugin._obj.max_sources = 1
    catalogs_plugin._obj.search(error_on_fail=True)
    assert catalogs_plugin._obj.number_of_results == catalogs_plugin._obj.max_sources

    catalogs_plugin.clear_table()

    # test single source edge case and docs recommended input file type
    sky_coord = SkyCoord(ra=337.5202807, dec=-20.83305528, unit='deg')
    tbl = Table({'sky_centroid': [sky_coord], 'label': ['Source_1']})
    n_entries = len(tbl)

    catalogs_plugin.import_catalog(tbl)
    out_tbl = catalogs_plugin._obj.search()
    assert len([out_tbl]) == n_entries
    assert catalogs_plugin._obj.number_of_results == n_entries
    assert len(imviz_helper.app.data_collection) == 2  # image + markers

    catalogs_plugin.clear_table()
    assert not catalogs_plugin._obj.results_available
    assert len(imviz_helper.app.data_collection) == 1  # markers gone for good

    assert imviz_helper.viewers['imviz-0']._obj.state.x_min == -0.5
    assert imviz_helper.viewers['imviz-0']._obj.state.x_max == 9.5
    assert imviz_helper.viewers['imviz-0']._obj.state.y_min == -0.5
    assert imviz_helper.viewers['imviz-0']._obj.state.y_max == 9.5
    # Re-populate the table with a new search
    out_tbl = catalogs_plugin._obj.search()
    assert len(out_tbl) > 0
    # Ensure at least one row is selected before zooming
    catalogs_plugin._obj.table.selected_rows = [catalogs_plugin._obj.table.items[0]]
    assert len(catalogs_plugin._obj.table.selected_rows) > 0

    # test the zooming using the default 'padding' of 2% of the viewer size
    # around selected points
    catalogs_plugin.zoom_to_selected()
    assert_allclose(imviz_helper.viewers['imviz-0']._obj.state.x_min, -0.01966, rtol=1e-4)
    assert_allclose(imviz_helper.viewers['imviz-0']._obj.state.x_max, 0.02034, rtol=1e-4)
    assert_allclose(imviz_helper.viewers['imviz-0']._obj.state.y_min, 0.980008, rtol=1e-4)
    assert_allclose(imviz_helper.viewers['imviz-0']._obj.state.y_max, 1.020008, rtol=1e-4)


def test_zoom_to_selected(imviz_helper, image_2d_wcs):
    # Since we are not really displaying, need this to test zoom.
    viewer = imviz_helper.default_viewer._obj
    viewer.shape = (100, 100)
    viewer.state._set_axes_aspect_ratio(1)

    arr = np.ones((500, 500))
    ndd = NDData(arr, wcs=image_2d_wcs)
    imviz_helper.load_data(ndd)

    # sources at pixel coords ~(100, 100), ~(200, 200)
    sky_coord = SkyCoord(ra=[337.49056532, 337.46086081],
                         dec=[-20.80555273, -20.7777673], unit='deg')
    tbl = Table({'sky_centroid': [sky_coord],
                 'label': ['Source_1', 'Source_2']})

    catalogs_plugin = imviz_helper.plugins['Catalog Search']
    catalogs_plugin.import_catalog(tbl)
    catalogs_plugin._obj.search()

    # select both sources
    catalogs_plugin.select_all()

    # check viewer limits before zoom
    xmin, xmax, ymin, ymax = imviz_helper.default_viewer._obj.get_limits()
    assert xmin == ymin == -0.5
    assert xmax == ymax == 499.5

    # zoom to selected sources
    catalogs_plugin.zoom_to_selected()

    # make sure the viewer bounds reflect the zoom, which, in pixel coords,
    # should be centered at roughly pixel coords (150, 150)
    xmin, xmax, ymin, ymax = imviz_helper.default_viewer._obj.get_limits()

    assert_allclose((xmin + xmax) / 2, 150., atol=0.1)
    assert_allclose((ymin + ymax) / 2, 150., atol=0.1)

    # and the zoom box size should reflect the default padding of 2% of the image
    # size around the bounding box containing the source(s), which in this case is
    # 10 pixels around
    assert_allclose(xmin, 96, atol=0.1)  # min x of selected sources minus pad
    assert_allclose(xmax, 204, atol=0.1)  # max x of selected sources plus pad
    assert_allclose(ymin, 96, atol=0.1)  # min y of selected sources minus pad
    assert_allclose(ymax, 204, atol=0.1)  # max y of selected sources plus pad

    # select one source now
    catalogs_plugin._obj.table.selected_rows = catalogs_plugin._obj.table.items[0:1]

    # zoom to single selected source, using a new value for 'padding'
    catalogs_plugin.zoom_to_selected(padding=0.05)

    # check that zoom window is centered correctly on the source at 100, 100
    xmin, xmax, ymin, ymax = imviz_helper.default_viewer._obj.get_limits()
    assert_allclose((xmin + xmax) / 2, 100., atol=0.1)
    assert_allclose((ymin + ymax) / 2, 100., atol=0.1)

    # and the zoom box size should reflect the default padding of 5% of the image
    # size around the bounding box containing the source(s), which in this case is
    # 25 pixels around
    assert_allclose(xmin, 95, atol=0.1)  # min x of selected source minus pad
    assert_allclose(xmax, 105, atol=0.1)  # max x of selected source plus pad
    assert_allclose(ymin, 95, atol=0.1)  # min y of selected source minus pad
    assert_allclose(ymax, 105, atol=0.1)  # max y of selected source plus pad

    # test that appropriate error is raised when padding is not a valud percentage
    with pytest.raises(ValueError, match="padding must be between 0 .* and 1.*"):
        catalogs_plugin.zoom_to_selected(padding=5)


def test_select_tool(imviz_helper, image_2d_wcs):
    arr = np.ones((500, 500))
    ndd = NDData(arr, wcs=image_2d_wcs)
    imviz_helper.load_data(ndd)

    # write out catalog to file so we can read it back in
    # todo: if tables can be loaded directly at some point, do that

    # sources at pixel coords ~(100, 100), ~(200, 200)
    sky_coord = SkyCoord(ra=[337.49056532, 337.46086081],
                         dec=[-20.80555273, -20.7777673], unit='deg')
    tbl = Table({'sky_centroid': [sky_coord],
                 'label': ['Source_1', 'Source_2']})

    catalogs_plugin = imviz_helper.plugins['Catalog Search']
    catalogs_plugin.import_catalog(tbl)

    # before catalog searching, the selection tool is not available
    toolbar = imviz_helper.viewers['imviz-0']._obj.toolbar
    tool = toolbar.tools['jdaviz:selectcatalog']
    mark = catalogs_plugin._obj._get_mark(imviz_helper.viewers['imviz-0']._obj)
    assert tool.is_visible() is False

    catalogs_plugin._obj.search(error_on_fail=True)
    assert tool.is_visible()

    assert len(catalogs_plugin._obj.table.selected_rows) == 0
    assert len(catalogs_plugin._obj.table_selected.items) == 0
    assert len(mark.x) == 0

    toolbar.active_tool_id = 'jdaviz:selectcatalog'
    tool.on_mouse_event({'domain': {'x': 100, 'y': 110}})

    assert len(catalogs_plugin._obj.table.selected_rows) == 1
    assert len(catalogs_plugin._obj.table_selected.items) == 1
    assert len(mark.x) == 1

    tool.on_mouse_event({'domain': {'x': 200, 'y': 210}})
    assert len(catalogs_plugin._obj.table.selected_rows) == 2
    assert len(catalogs_plugin._obj.table_selected.items) == 2
    assert len(mark.x) == 2

    # click to remove
    tool.on_mouse_event({'domain': {'x': 110, 'y': 110}})
    assert len(catalogs_plugin._obj.table.selected_rows) == 1
    assert len(catalogs_plugin._obj.table_selected.items) == 1
    assert len(mark.x) == 1

    catalogs_plugin._obj.table_selected.clear_table()
    assert len(catalogs_plugin._obj.table.selected_rows) == 0
    assert len(catalogs_plugin._obj.table_selected.items) == 0
    assert len(mark.x) == 0


def test_offline_ecsv_catalog_with_extra_columns(imviz_helper, image_2d_wcs):
    # Create a table with additional columns
    sky = SkyCoord(ra=[337.5202807, 337.51909197, 337.51760596],
                   dec=[-20.83305528, -20.83222194, -20.83083304], unit='deg')
    tbl = QTable({
        'sky_centroid': sky,
        'flux': [1.0, 2.0, 3.0],
        'flux_err': [0.1, 0.2, 0.3],
        'is_extended': [False, True, False],
        'roundness': [0.01, 0.02, 0.03],
        'sharpness': [0.1, 0.2, 0.3]
    })

    ndd = NDData(np.ones((10, 10)), wcs=image_2d_wcs)
    imviz_helper.load_data(ndd, data_label='data_with_wcs')
    assert len(imviz_helper.app.data_collection) == 1

    catalogs_plugin = imviz_helper.plugins['Catalog Search']._obj
    catalogs_plugin.import_catalog(tbl)
    catalogs_plugin.search(error_on_fail=True)

    extra_columns = ['flux', 'flux_err', 'is_extended', 'roundness', 'sharpness']
    for col in extra_columns:
        assert col in catalogs_plugin.table.headers_avail

    # Check if extra columns are populated correctly
    for idx, item in enumerate(catalogs_plugin.table.items):
        assert float(item['flux']) == tbl['flux'][idx]
        assert float(item['flux_err']) == tbl['flux_err'][idx]
        assert item['is_extended'] == tbl['is_extended'][idx]
        assert float(item['roundness']) == tbl['roundness'][idx]
        assert float(item['sharpness']) == tbl['sharpness'][idx]


def test_select_catalog_table_rows(imviz_helper, image_2d_wcs):
    """Test the ``select_rows`` functionality on table in plugin."""

    arr = np.ones((500, 500))
    ndd = NDData(arr, wcs=image_2d_wcs)
    imviz_helper.load_data(ndd)

    sky_coord = SkyCoord(ra=[337.49, 337.46, 337.47, 337.48, 337.49, 337.50],
                         dec=[-20.81, -20.78, -20.79, -20.80, -20.77, -20.76],
                         unit='deg')
    tbl = Table({'sky_centroid': [sky_coord],
                'label': ['Source_1', 'Source_2', 'Source_3', 'Source_4',
                          'Source_5', 'Source_6']})

    catalogs_plugin = imviz_helper.plugins['Catalog Search']
    plugin_table = catalogs_plugin._obj.table

    # load catalog
    catalogs_plugin.import_catalog(tbl)
    catalogs_plugin._obj.search()

    # select a single row:
    catalogs_plugin.select_rows(3)
    assert len(plugin_table.selected_rows) == 1
    assert plugin_table.selected_rows[0]['Right Ascension (degrees)'] == '337.48000'

    # select multiple rows by indices
    catalogs_plugin.select_rows([3, 2, 4])
    assert len(plugin_table.selected_rows) == 3
    assert plugin_table.selected_rows[0]['Right Ascension (degrees)'] == '337.48000'
    assert plugin_table.selected_rows[1]['Right Ascension (degrees)'] == '337.47000'
    assert plugin_table.selected_rows[2]['Right Ascension (degrees)'] == '337.49000'

    # select a range of rows with a slice
    catalogs_plugin.select_rows(slice(0, 2))
    assert len(plugin_table.selected_rows) == 2
    assert plugin_table.selected_rows[0]['Right Ascension (degrees)'] == '337.49000'
    assert plugin_table.selected_rows[1]['Right Ascension (degrees)'] == '337.46000'

    # select rows with multi dim. numpy slice
    catalogs_plugin.select_rows(np.s_[0:2, 3:5])
    assert len(plugin_table.selected_rows) == 4
    assert plugin_table.selected_rows[0]['Right Ascension (degrees)'] == '337.49000'
    assert plugin_table.selected_rows[1]['Right Ascension (degrees)'] == '337.46000'
    assert plugin_table.selected_rows[2]['Right Ascension (degrees)'] == '337.48000'
    assert plugin_table.selected_rows[3]['Right Ascension (degrees)'] == '337.49000'

    # test select_all
    catalogs_plugin.select_all()
    assert len(plugin_table.selected_rows) == 6

    # test select_none
    catalogs_plugin.select_none()
    assert len(plugin_table.selected_rows) == 0



from astropy import units as u
from astropy.wcs.wcsapi import BaseHighLevelWCS
from glue.core.link_helpers import LinkSame
from glue.core.message import (
    DataCollectionAddMessage, SubsetCreateMessage, SubsetDeleteMessage
)
from glue.core.subset import Subset
from glue.core.subset_group import GroupedSubset
from glue.plugins.wcs_autolinking.wcs_autolinking import WCSLink, NoAffineApproximation
from glue.viewers.image.state import ImageSubsetLayerState
from traitlets import List, Unicode, Bool, Dict, observe

from jdaviz.configs.imviz.wcs_utils import (
    get_compass_info, _get_rotated_nddata_from_label
)
from jdaviz.core.custom_traitlets import FloatHandleEmpty
from jdaviz.core.events import (
    ExitBatchLoadMessage, ChangeRefDataMessage,
    AstrowidgetMarkersChangedMessage, MarkersPluginUpdate,
    SnackbarMessage, ViewerAddedMessage, AddDataMessage, LinkUpdatedMessage
)
from jdaviz.core.registries import tray_registry
from jdaviz.core.template_mixin import (
    PluginTemplateMixin, SelectPluginComponent, LayerSelect, ViewerSelectMixin, AutoTextField
)
from jdaviz.core.user_api import PluginUserApi
from jdaviz.utils import (get_wcs_only_layer_labels, get_reference_image_data,
                          layer_is_2d, _wcs_only_label)

__all__ = ['Orientation']

orientation_plugin_label = "Orientation"
base_wcs_layer_label = 'Default orientation'
align_by_msg_to_trait = {'pixels': 'Pixels', 'wcs': 'WCS'}


@tray_registry('imviz-orientation', label=orientation_plugin_label, viewer_requirements="image")
class Orientation(PluginTemplateMixin, ViewerSelectMixin):
    """
    See the :ref:`Orientation Plugin Documentation <imviz-orientation>` for more details.

    .. note::
       Changing linking after adding markers via
       `~jdaviz.core.astrowidgets_api.AstrowidgetsImageViewerMixin.add_markers` is unsupported and
       will raise an error requiring resetting the markers manually via
       `~jdaviz.core.astrowidgets_api.AstrowidgetsImageViewerMixin.add_markers`
       or clicking a button in the plugin first.

    Only the following attributes and methods are available through the
    :ref:`public plugin API <plugin-apis>`:

    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.show`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.open_in_tray`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.close_in_tray`
    * ``align_by`` (`~jdaviz.core.template_mixin.SelectPluginComponent`)
    * ``wcs_fast_approximation``
    * :meth:`delete_subsets`
    * ``viewer``
    * ``orientation``
    * ``rotation_angle``
    * ``east_left``
    * :meth:`add_orientation`
    * :meth:`set_north_up_east_left`
    * :meth:`set_north_up_east_right`
    """
    template_file = __file__, "orientation.vue"

    # defined as traitlet in addition to global variable above to
    # allow access from UI - leave fixed
    base_wcs_layer_label = Unicode(base_wcs_layer_label).tag(sync=True)

    align_by_items = List().tag(sync=True)
    align_by_selected = Unicode().tag(sync=True)
    wcs_use_fallback = Bool(True).tag(sync=True)
    wcs_fast_approximation = Bool(True).tag(sync=True)
    wcs_linking_available = Bool(False).tag(sync=True)

    need_clear_astrowidget_markers = Bool(False).tag(sync=True)
    plugin_markers_exist = Bool(False).tag(sync=True)
    linking_in_progress = Bool(False).tag(sync=True)
    need_clear_subsets = Bool(False).tag(sync=True)

    # rotation angle, counterclockwise [degrees]
    rotation_angle = FloatHandleEmpty(0).tag(sync=True)
    east_left = Bool(True).tag(sync=True)  # set convention for east left of north

    icons = Dict().tag(sync=True)

    viewer_items = List().tag(sync=True)
    viewer_selected = Unicode().tag(sync=True)
    orientation_layer_items = List().tag(sync=True)
    orientation_layer_selected = Unicode().tag(sync=True)

    new_layer_label = Unicode().tag(sync=True)
    new_layer_label_default = Unicode().tag(sync=True)
    new_layer_label_auto = Bool(True).tag(sync=True)

    multiselect = Bool(False).tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # description displayed under plugin title in tray
        self._plugin_description = 'Rotate viewer orientation and choose alignment (pixel or sky).'

        self.icons = {k: v for k, v in self.app.state.icons.items()}

        self.align_by = SelectPluginComponent(self,
                                              items='align_by_items',
                                              selected='align_by_selected',
                                              manual_options=['Pixels', 'WCS'])

        self.orientation = LayerSelect(
            self, 'orientation_layer_items', 'orientation_layer_selected', 'viewer_selected',
            'multiselect', only_wcs_layers=True
        )
        self.orientation_layer_label = AutoTextField(
            self, 'new_layer_label', 'new_layer_label_default', 'new_layer_label_auto', None
        )

        self.hub.subscribe(self, DataCollectionAddMessage,
                           handler=self._on_new_app_data)

        self.hub.subscribe(self, ExitBatchLoadMessage,
                           handler=self._on_new_app_data)

        self.hub.subscribe(self, AstrowidgetMarkersChangedMessage,
                           handler=self._on_astrowidget_markers_changed)

        self.hub.subscribe(self, MarkersPluginUpdate,
                           handler=self._on_markers_plugin_update)

        self.hub.subscribe(self, ChangeRefDataMessage,
                           handler=self._on_refdata_change)

        self.hub.subscribe(self, SubsetCreateMessage,
                           handler=self._on_subset_change)

        self.hub.subscribe(self, SubsetDeleteMessage,
                           handler=self._on_subset_change)

        self.hub.subscribe(self, ViewerAddedMessage,
                           handler=self._on_viewer_added)

        self.hub.subscribe(self, AddDataMessage,
                           handler=self._on_data_add_to_viewer)

        self._update_layer_label_default()

    @property
    def user_api(self):
        return PluginUserApi(
            self,
            expose=(
                'align_by', 'wcs_fast_approximation',
                'delete_subsets', 'viewer', 'orientation',
                'rotation_angle', 'east_left', 'add_orientation',
                'set_north_up_east_left', 'set_north_up_east_right',
            )
        )

    def _link_image_data(self):
        self.linking_in_progress = True
        try:
            align_by = self.align_by.selected.lower()
            link_image_data(
                self.app,
                align_by=align_by,
                wcs_fallback_scheme='pixels' if self.wcs_use_fallback else None,
                wcs_fast_approximation=self.wcs_fast_approximation,
                error_on_fail=False)
        except Exception:  # pragma: no cover
            raise
        else:
            # Only broadcast after success.
            self.app.hub.broadcast(LinkUpdatedMessage(
                align_by, self.wcs_use_fallback, self.wcs_fast_approximation, sender=self.app))
        finally:
            self.linking_in_progress = False

    def _check_if_data_with_wcs_exists(self):
        for data in self.app.data_collection:
            if hasattr(data.coords, 'pixel_to_world'):
                self.wcs_linking_available = True
                return
        self.wcs_linking_available = False

    def _on_new_app_data(self, msg):
        if self.app._jdaviz_helper._in_batch_load > 0:
            return
        if isinstance(msg, DataCollectionAddMessage):
            components = [str(comp) for comp in msg.data.main_components]
            if "ra" in components or "Lon" in components:
                # linking currently removes any markers, so we want to skip
                # linking immediately after new markers are added.
                # Eventually we'll probably want to support linking WITH markers,
                # at which point this if-statement should be removed.
                return
        self._link_image_data()
        self._check_if_data_with_wcs_exists()

    def _on_astrowidget_markers_changed(self, msg):
        self.need_clear_astrowidget_markers = msg.has_markers

    def _on_markers_plugin_update(self, msg):
        self.plugin_markers_exist = msg.table_length > 0

    @observe('align_by_selected', 'wcs_use_fallback', 'wcs_fast_approximation')
    def _update_link(self, msg={}):
        """Run link_image_data with the selected parameters."""
        if not hasattr(self, 'align_by'):
            # could happen before plugin is fully initialized
            return

        if msg.get('name', None) == 'wcs_fast_approximation' and self.align_by.selected == 'Pixels':  # noqa
            # approximation doesn't apply, avoid updating when not necessary!
            return

        if self.linking_in_progress:
            return

        self.linking_in_progress = True  # Prevent recursion

        if self.need_clear_subsets:
            self.linking_in_progress = False
            raise ValueError("Link type can only be changed after existing subsets "
                             f"are deleted, but {len(self.app.data_collection.subset_groups)} "
                             f"subset(s) still exist. To delete them, you can use "
                             f"`delete_subsets()` from the plugin API.")

        if self.need_clear_astrowidget_markers:
            setattr(self, msg.get('name'), msg.get('old'))
            self.linking_in_progress = False
            raise ValueError(f"cannot change linking with markers present (value reverted to "
                             f"'{msg.get('old')}'), call viewer.reset_markers()")

        if self.align_by.selected == 'Pixels':
            self.wcs_fast_approximation = True

        self._link_image_data()
        # NOTE: _link_image_data will reset linking_in_progress to False

        # load data into the viewer that are now compatible with the
        # new link type, remove data from the viewer that are now
        # incompatible:
        wcs_linked = self.align_by.selected == 'WCS'
        viewer_selected = self.app.get_viewer(self.viewer.selected)

        data_in_viewer = self.app.get_viewer(viewer_selected.reference).data()

        for data in self.app.data_collection:
            is_wcs_only = data.meta.get(_wcs_only_label, False)
            has_wcs = hasattr(data.coords, 'pixel_to_world')
            if not is_wcs_only:
                if data in data_in_viewer and wcs_linked and not has_wcs:
                    # data is in viewer but must be removed:
                    self.app.remove_data_from_viewer(viewer_selected.reference, data.label)
                    self.hub.broadcast(SnackbarMessage(
                        f"Data '{data.label}' does not have a valid WCS - removing from viewer.",
                        sender=self, color="warning"))

        if wcs_linked:
            self._send_wcs_layers_to_all_viewers()

        self._update_layer_label_default()

        # Clear previous zoom limits because they no longer mean anything.
        for v in self.app._viewer_store.values():
            v._prev_limits = None

    def _on_subset_change(self, msg):
        self.need_clear_subsets = len(self.app.data_collection.subset_groups) > 0

    def delete_subsets(self):
        """
        Delete all subsets app-wide.  Required before changing ``align_by``.
        """
        # subsets will be deleted on changing link type:
        self.app.delete_subsets()

    def vue_delete_subsets(self, *args):
        self.delete_subsets()

    def vue_reset_astrowidget_markers(self, *args):
        for viewer in self.app._viewer_store.values():
            viewer.reset_markers()

    def _get_wcs_angles(self, first_loaded_image=None):
        if first_loaded_image is None:
            first_loaded_image = self.viewer.selected_obj.first_loaded_data
            if first_loaded_image is None:
                # These won't end up getting used in this case, but we need an actual number
                return 0, 0, 0
        degn, dege, flip = get_compass_info(
            first_loaded_image.coords, first_loaded_image.shape
        )[-3:]
        return degn, dege, flip

    def rotation_angle_deg(self, rotation_angle=None):
        if rotation_angle is None:
            rotation_angle = self.rotation_angle

        if rotation_angle is not None:
            if (
                (isinstance(rotation_angle, str) and len(rotation_angle)) or
                isinstance(rotation_angle, (int, float))
            ):
                return float(rotation_angle) * u.deg
        return 0 * u.deg

    def add_orientation(self, rotation_angle=None, east_left=None, label=None,
                        set_on_create=True, wrt_data=None):
        """
        Add new orientation options.

        Parameters
        ----------
        rotation_angle : float, optional
            Desired sky orientation angle in degrees.
            If `None`, the value will follow ``self.rotation_angle``.
            If nothing is set anywhere, it defaults to zero degrees.

        east_left : bool, optional
            Set to `True` if you want N-up E-left or `False` for N-up E-right.
            If `None`, the value will follow ``self.east_left``.

        label : str, optional
            Data label for this new orientation layer.
            If `None`, the value will follow ``self.new_layer_label``.

        set_on_create : bool, optional
            If `True`, this new orientation layer will become active
            on creation. Otherwise, it will be created but stay inactive
            in the background.

        wrt_data : str, optional
            Orientation calculations is done with respect to this data WCS.
            If `None`, it grabs the first loaded data in the selected viewer
            (may or may not be visible); If no data is loaded in the viewer,
            nothing will be done.

        """
        self._add_orientation(rotation_angle=rotation_angle, east_left=east_left, label=label,
                              set_on_create=set_on_create, wrt_data=wrt_data)

    def _add_orientation(self, rotation_angle=None, east_left=None, label=None,
                         set_on_create=True, wrt_data=None, from_ui=False):

        if self.align_by_selected != 'WCS':
            raise ValueError("must be aligned by WCS to add orientation options")

        if wrt_data is None:
            # if not specified, use first-loaded image layer as the
            # default rotation:
            wrt_data = self.viewer.selected_obj.first_loaded_data
            if wrt_data is None:  # Nothing in viewer
                msg = "Viewer must have data loaded to add an orientation."
                if from_ui:
                    self.hub.broadcast(SnackbarMessage(msg, color="error",
                                                       timeout=6000, sender=self))
                else:
                    raise ValueError(msg)
                return

        rotation_angle = self.rotation_angle_deg(rotation_angle)
        if east_left is None:
            east_left = self.east_left
        if label is None:
            label = self.new_layer_label

        # Default rotation is the same orientation as the original reference data:
        degn = self._get_wcs_angles(first_loaded_image=wrt_data)[0]

        if east_left:
            rotation_angle = -degn * u.deg + rotation_angle
        else:
            rotation_angle = (180 - degn) * u.deg - rotation_angle

        ndd = _get_rotated_nddata_from_label(
            app=self.app,
            data_label=wrt_data.label,
            rotation_angle=rotation_angle,
            target_wcs_east_left=east_left,
            target_wcs_north_up=True,
        )

        self.app._jdaviz_helper.load_data(
            ndd, data_label=label
        )

        # add orientation layer to all viewers:
        for viewer_ref in self.app._viewer_store:
            self._add_data_to_viewer(label, viewer_ref)

        if set_on_create:
            # Not sure why this is sometimes missing, but we can add it here
            if label not in self.orientation.choices:
                self.orientation.choices += [label]
            self.orientation.selected = label

    def _add_data_to_viewer(self, data_label, viewer_id):
        viewer = self.app.get_viewer_by_id(viewer_id)

        if data_label not in viewer.data_menu.orientation.choices:
            self.app.add_data_to_viewer(viewer_id, data_label)

    def _on_viewer_added(self, msg):
        self._send_wcs_layers_to_all_viewers(viewers_to_update=[msg._viewer_id])

    @observe('viewer_items')
    def _send_wcs_layers_to_all_viewers(self, *args, **kwargs):
        if not hasattr(self, 'viewer'):
            return

        wcs_only_layers = get_wcs_only_layer_labels(self.app)

        viewers_to_update = kwargs.get(
            'viewers_to_update', self.app._viewer_store.keys()
        )
        for viewer_ref in viewers_to_update:
            viewer_dm = self.app._jdaviz_helper.viewers.get(viewer_ref).data_menu
            for wcs_layer in wcs_only_layers:
                if wcs_layer not in self.viewer.selected_obj.layers:
                    self.app.add_data_to_viewer(viewer_ref, wcs_layer)
            if (
                self.orientation.selected not in
                    wcs_only_layers and
                    self.align_by_selected == 'WCS'
            ):
                viewer_dm.orientation.selected = base_wcs_layer_label

    def _on_data_add_to_viewer(self, msg):
        all_wcs_only_layers = all(
            layer.layer.meta.get(_wcs_only_label)
            for layer in self.viewer.selected_obj.layers
            if hasattr(layer.layer, "meta")
        )
        if all_wcs_only_layers and msg.data.meta.get(_wcs_only_label, False):
            # on adding first data layer, reset the limits:
            self.viewer.selected_obj.state.reset_limits()

    def vue_add_orientation(self, *args, **kwargs):
        self._add_orientation(set_on_create=True, from_ui=True)

    def set_orientation_for_viewer(self, orientation, viewer_id):
        if self._refdata_change_available:
            self.app._change_reference_data(
                orientation, viewer_id=viewer_id
            )
            viewer_item = self.app._viewer_item_by_id(viewer_id)
            if viewer_item['reference_data_label'] != orientation:
                viewer_item['reference_data_label'] = orientation

    @observe('orientation_layer_selected')
    def _change_reference_data(self, *args, **kwargs):
        self.set_orientation_for_viewer(self.orientation.selected, self.viewer.selected)

    def _on_refdata_change(self, msg):

        if hasattr(self, 'viewer'):
            ref_data = self.ref_data
            viewer = self.viewer.selected_obj

            # don't select until reference data are available:
            if ref_data is not None:
                align_by = viewer.get_alignment_method(ref_data.label)
                if align_by != 'self':
                    self.align_by_selected = align_by_msg_to_trait[align_by]
            elif not len(viewer.data()):
                self.align_by_selected = align_by_msg_to_trait['pixels']

            if msg.data.label not in self.orientation.choices:
                return

            if msg.viewer_id == self.viewer.selected:
                self.orientation.selected = msg.data.label

            # we never want to highlight subsets of pixels within WCS-only layers,
            # so if this layer is an ImageSubsetLayerState on a WCS-only layer,
            # ensure that it is never visible:
            for layer in viewer.state.layers:
                if (
                    isinstance(layer.layer, ImageSubsetLayerState) and
                    layer.layer.data.meta.get("_WCS_ONLY", False)
                ):
                    layer.visible = False

    @property
    def ref_data(self):
        if hasattr(self, 'viewer'):
            return self.app.get_viewer_by_id(self.viewer.selected).state.reference_data
        return None

    @property
    def _refdata_change_available(self):
        viewer = self.app.get_viewer(self.viewer.selected)
        selected_layer = [lyr.layer for lyr in viewer.layers
                          if lyr.layer.label == self.orientation.selected]
        if len(selected_layer):
            is_subset = isinstance(selected_layer[0], (Subset, GroupedSubset))
        else:
            is_subset = False
        return (
            len(self.orientation.selected) and
            len(self.viewer.selected) and
            not is_subset
        )

    @observe('viewer_selected')
    def _on_viewer_change(self, msg={}):
        # don't update choices until viewer is available:
        ref_data = self.ref_data
        if hasattr(self, 'viewer') and ref_data is not None:
            if ref_data.label in self.orientation.choices:
                self.orientation.selected = ref_data.label

    def _set_north_up_east_left(self, label="North-up, East-left", set_as_orientation=False,
                                from_ui=False):
        if label not in self.orientation.choices:
            degn = self._get_wcs_angles()[-3]
            self._add_orientation(rotation_angle=degn, east_left=True,
                                  label=label, set_on_create=set_as_orientation,
                                  from_ui=from_ui)
        elif set_as_orientation:
            self.orientation.selected = label

    def set_north_up_east_left(self, label="North-up, East-left"):
        """
        Set (and create if necessary) the rotation angle and flip to achieve North up
        and East left according to the reference image WCS.

        Parameters
        ----------
        label : str
            Data label for the orientation layer.  If already exists, will be set as the
            current orientation layer according to ``set_as_orientation``.  Otherwise,
            a new layer will be created with this label.
        """
        self._set_north_up_east_left(label=label, set_as_orientation=True)

    def _set_north_up_east_right(self, label="North-up, East-right", set_as_orientation=False,
                                 from_ui=False):
        if label not in self.orientation.choices:
            degn = self._get_wcs_angles()[-3]
            self._add_orientation(rotation_angle=180 - degn, east_left=False,
                                  label=label, set_on_create=set_as_orientation,
                                  from_ui=from_ui)
        elif set_as_orientation:
            self.orientation.selected = label

    def set_north_up_east_right(self, label="North-up, East-right"):
        """
        Set (and create, if necessary) the rotation angle and flip to achieve North up
        and East right according to the reference image WCS.

        Parameters
        ----------
        label : str
            Data label for the orientation layer.  If already exists, will be set as the
            current orientation layer according to ``set_as_orientation``.  Otherwise,
            a new layer will be created with this label.
        """
        self._set_north_up_east_right(label=label, set_as_orientation=True)

    def vue_select_north_up_east_left(self, *args, **kwargs):
        self._set_north_up_east_left(set_as_orientation=True, from_ui=True)

    def vue_select_north_up_east_right(self, *args, **kwargs):
        self._set_north_up_east_right(set_as_orientation=True, from_ui=True)

    def vue_select_default_orientation(self, *args, **kwargs):
        self.orientation.selected = base_wcs_layer_label

    @observe('east_left', 'rotation_angle')
    def _update_layer_label_default(self, event={}):
        self.new_layer_label_default = (
            f'CCW {self.rotation_angle_deg():.2f} ' +
            ('(E-left)' if self.east_left else '(E-right)')
        )


def link_image_data(app, align_by='pixels', wcs_fallback_scheme=None, wcs_fast_approximation=True,
                    error_on_fail=False):
    """(Re)link loaded data in Imviz with the desired link type.

    .. note::

        Any markers added in Imviz will need to be removed manually before changing linking type.
        You can add back the markers using
        :meth:`~jdaviz.core.astrowidgets_api.AstrowidgetsImageViewerMixin.add_markers`
        for the relevant viewer(s).

    Parameters
    ----------
    app : `~jdaviz.app.Application`
        Application associated with Imviz, e.g., ``imviz.app``.

    align_by : {'pixels', 'wcs'}
        Choose to link by pixels or WCS.

    wcs_fallback_scheme : {None, 'pixels'}
        If WCS linking failed, choose to fall back to linking by pixels or not at all.
        This is only used when ``align_by='wcs'``.
        Choosing `None` may result in some Imviz functionality not working properly.

    wcs_fast_approximation : bool
        Use an affine transform to represent the offset between images if possible
        (requires that the approximation is accurate to within 1 pixel with the
        full WCS transformations). If approximation fails, it will automatically
        fall back to full WCS transformation. This is only used when ``align_by='wcs'``.
        Affine approximation is much more performant at the cost of accuracy.

    error_on_fail : bool
        If `True`, any failure in linking will raise an exception.
        If `False`, warnings will be emitted as snackbar messages.
        When only warnings are emitted and no links are assigned,
        some Imviz functionality may not work properly.

    Raises
    ------
    ValueError
        Invalid inputs or reference data.

    """
    if len(app.data_collection) <= 1 and align_by != 'wcs':  # No need to link, we are done.
        return

    if align_by not in ('pixels', 'wcs'):  # pragma: no cover
        raise ValueError(f"align_by must be 'pixels' or 'wcs', got {align_by}")
    if align_by == 'wcs' and wcs_fallback_scheme not in (None, 'pixels'):  # pragma: no cover
        raise ValueError("wcs_fallback_scheme must be None or 'pixels', "
                         f"got {wcs_fallback_scheme}")
    if align_by == 'wcs':
        at_least_one_data_have_wcs = len([
            hasattr(d, 'coords') and isinstance(d.coords, BaseHighLevelWCS)
            for d in app.data_collection
        ]) > 0
        if not at_least_one_data_have_wcs:  # pragma: no cover
            if wcs_fallback_scheme is None:
                if error_on_fail:
                    raise ValueError("align_by can only be 'wcs' when wcs_fallback_scheme "
                                     "is 'None' if at least one image has a valid WCS.")
                else:
                    return
            else:
                # fall back on pixel linking
                align_by = 'pixels'

    old_align_by = getattr(app, '_align_by', None)

    # In WCS linking, changing orientation layer is done within Orientation plugin,
    # so here we assume viewer.state.reference_data is already the desired
    # orientation by the time this function is called.
    #
    # In pixels linking, Affine approximation does not matter.
    #
    # data1 = reference, data2 = actual data
    data_already_linked = []
    if (align_by == old_align_by and
            (align_by == "pixels" or wcs_fast_approximation == app._wcs_fast_approximation)):
        # We are only here to link new data with existing configuration,
        # so no need to relink existing data.
        for link in app.data_collection.external_links:
            data_already_linked.append(link.data2)
    else:  # pragma: no cover
        # Everything has to be relinked.
        for viewer in app._viewer_store.values():
            if len(viewer._marktags):
                raise ValueError(f"cannot change align_by (from '{app._align_by}' to "
                                 f"'{align_by}') when markers are present. "
                                 f" Clear markers with viewer.reset_markers() first")

    # set internal tracking of align_by before changing reference data for anything that is
    # subscribed to a change in reference data
    app._align_by = align_by
    app._wcs_fast_approximation = wcs_fast_approximation

    # wcs -> pixels: First loaded real data will be reference.
    if align_by == 'pixels' and old_align_by == 'wcs':
        # default reference layer is the first-loaded image in default viewer:
        refdata = app._jdaviz_helper.default_viewer._obj.first_loaded_data
        if refdata is None:  # No data in viewer, just use first in collection  # pragma: no cover
            iref = 0
            refdata = app.data_collection[iref]
        else:
            iref = app.data_collection.index(refdata)

        # set default layer to reference data in all viewers:
        for viewer_id in app.get_viewer_ids():
            app._change_reference_data(refdata.label, viewer_id=viewer_id)

    # pixels -> wcs: Always the default orientation
    elif align_by == 'wcs' and old_align_by == 'pixels':
        # Have to create the default orientation first.
        if base_wcs_layer_label not in app.data_collection.labels:
            default_reference_layer = (app._jdaviz_helper.default_viewer._obj.first_loaded_data
                                       or app.data_collection[0])
            degn = get_compass_info(default_reference_layer.coords, default_reference_layer.shape)[-3]  # noqa: E501
            # Default rotation is the same orientation as the original reference data:
            rotation_angle = -degn * u.deg
            ndd = _get_rotated_nddata_from_label(app, default_reference_layer.label, rotation_angle)
            app._jdaviz_helper.load_data(ndd, base_wcs_layer_label)

        # set default layer to reference data in all viewers:
        for viewer_id in app.get_viewer_ids():
            app._change_reference_data(base_wcs_layer_label, viewer_id=viewer_id)

        refdata = app.data_collection[base_wcs_layer_label]
        iref = app.data_collection.index(refdata)

    # Re-use current reference data.
    else:
        refdata, iref = get_reference_image_data(app)
        # App just loaded, nothing yet, so take first image.
        if refdata is None:
            refdata = app._jdaviz_helper.default_viewer._obj.first_loaded_data
            if refdata is None:  # No data in viewer, just use first in collection
                iref = 0
                refdata = app.data_collection[iref]
            else:   # pragma: no cover
                iref = app.data_collection.index(refdata)

    # With reference data changed, if needed, now we relink as needed.

    links_list = []
    ids0 = refdata.pixel_component_ids
    ndim_range = range(2)  # We only support 2D

    for i, data in enumerate(app.data_collection):
        # 1. Do not link with self.
        # 2. We are not touching any existing Subsets or Table. They keep their own links.
        # 3. Links already exist for this entry and we're not changing the type.
        # 4. We are not touching fake WCS layers in pixel linking.
        # 5. We are not touching data without WCS in WCS linking.
        if ((i == iref) or (not layer_is_2d(data)) or (data in data_already_linked) or
                (align_by == "pixels" and data.meta.get(_wcs_only_label)) or
                (align_by == "wcs" and not hasattr(data.coords, 'pixel_to_world'))):
            continue

        ids1 = data.pixel_component_ids
        new_links = []
        try:
            if align_by == 'pixels':
                new_links = [LinkSame(ids0[i], ids1[i]) for i in ndim_range]
            else:  # wcs
                wcslink = WCSLink(data1=refdata, data2=data, cids1=ids0, cids2=ids1)
                if wcs_fast_approximation:
                    try:
                        new_links = [wcslink.as_affine_link()]
                    except NoAffineApproximation:  # pragma: no cover
                        new_links = [wcslink]
                else:
                    new_links = [wcslink]
        except Exception as e:  # pragma: no cover
            if align_by == 'wcs' and wcs_fallback_scheme == 'pixels':
                try:
                    new_links = [LinkSame(ids0[i], ids1[i]) for i in ndim_range]
                except Exception as e:  # pragma: no cover
                    if error_on_fail:
                        raise
                    else:
                        app.hub.broadcast(SnackbarMessage(
                            f"Error linking '{data.label}' to '{refdata.label}': "
                            f"{repr(e)}", color="warning", timeout=8000, sender=app))
                        continue
            else:  # pragma: no cover
                if error_on_fail:
                    raise
                else:
                    app.hub.broadcast(SnackbarMessage(
                        f"Error linking '{data.label}' to '{refdata.label}': "
                        f"{repr(e)}", color="warning", timeout=8000, sender=app))
                    continue
        links_list += new_links

    if len(links_list) > 0:
        with app.data_collection.delay_link_manager_update():
            if len(data_already_linked):
                app.data_collection.add_link(links_list)  # Add to existing
            else:
                app.data_collection.set_links(links_list)  # Redo all links

        app.hub.broadcast(SnackbarMessage(
            'Images successfully relinked', color='success', timeout=8000, sender=app))

    for viewer in app._viewer_store.values():
        wcs_linked = align_by == 'wcs'
        # viewer-state needs to know link type for reset_limits behavior
        viewer.state.linked_by_wcs = wcs_linked
        # also need to store a copy in the viewer item for the data dropdown to access
        viewer_item = app._get_viewer_item(viewer.reference)

        viewer_item['reference_data_label'] = refdata.label
        viewer_item['linked_by_wcs'] = wcs_linked

        # if changing from one link type to another, reset the limits:
        if align_by != old_align_by:
            viewer.state.reset_limits()



from jdaviz import Application


def test_default_viewer_ids_default():
    app = Application(configuration='default')
    assert app.get_viewer_reference_names() == []
    assert app.get_viewer_ids() == []


# NOTE: Unable to parametrize fixture as input.

def test_default_viewer_ids_cubeviz(cubeviz_helper):
    x = cubeviz_helper.app
    assert sorted(x.get_viewer_reference_names()) == [
        'flux-viewer', 'spectrum-viewer', 'uncert-viewer']
    assert x.get_viewer_ids() == ['cubeviz-0', 'cubeviz-1', 'cubeviz-2']


def test_default_viewer_ids_imviz(imviz_helper):
    x = imviz_helper.app
    assert x.get_viewer_reference_names() == ['imviz-0']
    assert x.get_viewer_ids() == ['imviz-0']
    assert x.get_viewer_ids(prefix='imviz') == ['imviz-0']
    assert x.get_viewer_ids(prefix='specviz') == []

    viewer = x.get_viewer('imviz-0')
    assert x.get_viewer_by_id('imviz-0') is viewer


def test_default_viewer_ids_mosviz(mosviz_helper):
    x = mosviz_helper.app
    assert x.get_viewer_reference_names() == ['image-viewer', 'spectrum-2d-viewer',
                                              'spectrum-viewer', 'table-viewer']
    assert x.get_viewer_ids() == ['mosviz-0', 'mosviz-1', 'mosviz-2', 'mosviz-3']


def test_default_viewer_ids_specviz(specviz_helper):
    x = specviz_helper.app
    assert x.get_viewer_reference_names() == ['spectrum-viewer']
    assert x.get_viewer_ids() == ['specviz-0']


def test_default_viewer_ids_specviz2d(specviz2d_helper):
    x = specviz2d_helper.app
    assert x.get_viewer_reference_names() == ['spectrum-2d-viewer', 'spectrum-viewer']
    assert x.get_viewer_ids() == ['specviz2d-0', 'specviz2d-1']



from astropy.io import fits
import numpy as np
import pytest

from jdaviz.configs.imviz.tests.utils import BaseImviz_WCS_WCS
from jdaviz.configs.default.plugins.virtual_observatory.vo_plugin import vo_plugin_label


class fake_siaresult:
    """A mock class that simulates a SIAResult"""

    def __init__(self, attrs):
        for attr, value in attrs.items():
            self.__setattr__(attr, value)

    def getdataurl(self):
        return "Fake URL"


# TODO: Update all _obj calls to formal API calls once Plugin API is available
class TestVOImvizLocal(BaseImviz_WCS_WCS):
    _data_center_coords = {
        "has_wcs_1[SCI,1]": {"ra": 337.51894336761296, "dec": -20.832083054811765},
        "has_wcs_2[SCI,1]": {"ra": 337.51924057481, "dec": -20.83208305686149},
    }

    def test_autocenter_coords(self):
        """
        Loads two data products and checks the plugin correctly adjusts the automatically-derived
        coordinates of the center of the viewer when the viewer dropdown is changed.

        Also verify changing autocoord to a blank viewer with no data properly empties the source
        field.
        """
        # Create a second viewer and remove second dataset from first viewer to avoid ambiguity
        self.imviz.create_image_viewer()
        self.imviz.app.remove_data_from_viewer("imviz-0", "has_wcs_2[SCI,1]")

        # Check default viewer is "Manual"
        vo_plugin = self.imviz.plugins[vo_plugin_label]._obj
        assert vo_plugin.viewer.selected == "Manual"
        assert vo_plugin.source == ""

        # Switch to first viewer and verify coordinates have switched
        vo_plugin.viewer.selected = "imviz-0"
        ra_str, dec_str = vo_plugin.source.split()
        np.testing.assert_allclose(
            float(ra_str), self._data_center_coords["has_wcs_1[SCI,1]"]["ra"]
        )
        np.testing.assert_allclose(
            float(dec_str), self._data_center_coords["has_wcs_1[SCI,1]"]["dec"]
        )

        # Switch to second viewer without data and verify autocoord gracefully clears source field
        vo_plugin.viewer.selected = "imviz-1"
        assert vo_plugin.source == ""

        # Now load second data into second viewer and verify coordinates
        self.imviz.app.add_data_to_viewer("imviz-1", "has_wcs_2[SCI,1]")
        ra_str, dec_str = vo_plugin.source.split()
        np.testing.assert_allclose(
            float(ra_str), self._data_center_coords["has_wcs_2[SCI,1]"]["ra"]
        )
        np.testing.assert_allclose(
            float(dec_str), self._data_center_coords["has_wcs_2[SCI,1]"]["dec"]
        )

    def test_populate_table_default_headers(self):
        """
        Tests populating the results table with a mocked SIARsult
        and verifies the default headers populate the correct fields
        """
        vo_plugin = self.imviz.plugins[vo_plugin_label]._obj
        fake_result = fake_siaresult(
            {
                "title": "Fake Title",
                "instr": "Fake Instrument",
                "dateobs": "Fake Dateobs",
            }
        )
        vo_plugin._populate_table([fake_result])

        assert vo_plugin.table.items == [
            {
                "URL": fake_result.getdataurl(),
                "Title": fake_result.title,
                "Instrument": fake_result.instr,
                "DateObs": fake_result.dateobs,
            }
        ]

    def test_populate_table_custom_headers(self):
        """Tests the ability to control and adjust the table to custom columns"""
        vo_plugin = self.imviz.plugins[vo_plugin_label]._obj
        fake_result = fake_siaresult(
            {
                "attrA": "Field A",
                "attrB": "Field B",
                "attrC": "Field C",
            }
        )
        vo_plugin._populate_table(
            [fake_result], {"Title A": "attrA", "Title B": "attrB", "Title C": "attrC"}
        )

        assert vo_plugin.table.items == [
            {
                "URL": fake_result.getdataurl(),
                "Title A": fake_result.attrA,
                "Title B": fake_result.attrB,
                "Title C": fake_result.attrC,
            }
        ]

    def test_populate_table_url_header_fallback(self):
        """
        If a SIAResult is missing a required header,
        the table should switch to displaying URLs only
        """
        vo_plugin = self.imviz.plugins[vo_plugin_label]._obj
        fake_result = fake_siaresult(
            {
                "attrA": "Field A",
                "attrC": "Field C",
            }
        )

        vo_plugin._populate_table(
            [fake_result], {"Title A": "attrA", "Title B": "attrB", "Title C": "attrC"}
        )
        assert vo_plugin._populate_url_only is True
        assert vo_plugin.table.headers_visible == ["URL"]


def test_link_type_autocoord(imviz_helper):
    """
    Tests switching linking types forces recalculation of viewer center coordinates
    """
    # First data with WCS, same as the one in BaseImviz_WCS_NoWCS.
    hdu1 = fits.ImageHDU(np.random.rand(100, 100), name="SCI")
    hdu1.header.update(
        {
            "CTYPE1": "RA---TAN",
            "CUNIT1": "deg",
            "CDELT1": -2.777777778,
            "CRPIX1": 1,
            "CRVAL1": 337.5202808,
            "NAXIS1": 10,
            "CTYPE2": "DEC--TAN",
            "CUNIT2": "deg",
            "CDELT2": 2.777777778,
            "CRPIX2": 1,
            "CRVAL2": -20.833333059999998,
            "NAXIS2": 10,
        }
    )
    imviz_helper.load_data(hdu1, data_label="has_wcs_1")

    # Second data with WCS, similar to above but dithered by 1 pixel in X.
    hdu2 = fits.ImageHDU(np.ones((10, 10)), name="SCI")
    hdu2.header.update(
        {
            "CTYPE1": "RA---TAN",
            "CUNIT1": "deg",
            "CDELT1": -0.0002777777778,
            "CRPIX1": 2,
            "CRVAL1": 137.5202808,
            "NAXIS1": 10,
            "CTYPE2": "DEC--TAN",
            "CUNIT2": "deg",
            "CDELT2": 0.0002777777778,
            "CRPIX2": 1,
            "CRVAL2": -20.833333059999998,
            "NAXIS2": 10,
        }
    )
    imviz_helper.load_data(hdu2, data_label="has_wcs_2")

    vo_plugin = imviz_helper.plugins[vo_plugin_label]._obj
    vo_plugin.viewer.selected = "imviz-0"
    vo_plugin.center_on_data()
    ra_str, dec_str = vo_plugin.source.split()
    np.testing.assert_allclose(float(ra_str), 284.2101962057667)
    np.testing.assert_allclose(float(dec_str), 32.23616603681311)

    imviz_helper.plugins["Orientation"].align_by = "WCS"

    ra_str, dec_str = vo_plugin.source.split()

    # Large absolute tolerances due to WCS center coordinate bug (see issue 3225)
    # Truth values may need to be reevaluated
    np.testing.assert_allclose(float(ra_str), 326.7884142245305, atol=30)
    np.testing.assert_allclose(float(dec_str), -9.905948925234416, atol=30)


@pytest.mark.remote_data
class TestVOImvizRemote:

    def _init_voplugin_M51(self, imviz_helper):
        """
        Initialize vo plugin with common test parameters

        Parameters
        ----------
        imviz_helper : `~jdaviz.configs.imviz.helper`
            Instance of Imviz in which to initialize the VO plugin

        Returns
        -------
        vo_plugin_api : `~jdaviz.configs.default.plugins.virtual_observatory.VoPlugin`
            The raw VoPlugin class plugin instance
        """
        vo_plugin = imviz_helper.plugins[vo_plugin_label]._obj

        # Sets common args for Remote Testing
        vo_plugin.viewer.selected = "Manual"
        vo_plugin.source = "M51"
        vo_plugin.radius = 1
        vo_plugin.radius_unit.selected = "deg"
        vo_plugin.waveband.selected = "optical"

        return vo_plugin

    def test_query_registry_args(self, imviz_helper):
        """Ensure we don't query registry if we're missing required arguments"""
        # If waveband isn't selected, plugin should ignore our registry query attempts
        vo_plugin = imviz_helper.plugins[vo_plugin_label]._obj
        vo_plugin.waveband.selected = ""
        vo_plugin.query_registry_resources()
        assert len(vo_plugin.resource.choices) == 0

        # If waveband selected and coverage filtering, can't query registry if don't have a source
        vo_plugin.resource_filter_coverage = True
        vo_plugin.source = ""
        with pytest.raises(
            ValueError,
            match="Source is required for registry querying when coverage filtering is enabled.",
        ):
            # Setting the waveband from nothing to something will trigger the query
            vo_plugin.waveband.selected = "optical"
            # Also verify we get a snackbar message for it
            assert any(
                "Source is required" in d["text"]
                for d in imviz_helper.app.state.snackbar_history
            )

        # If waveband selected, but NOT filtering by coverage, then allow registry query
        vo_plugin.resource_filter_coverage = False
        vo_plugin.query_registry_resources()
        assert len(vo_plugin.resource.choices) > 0

    def test_coverage_toggle(self, imviz_helper):
        """
        Test that disabling the coverage toggle returns more available services

        NOTE: This does assume there exists at least one survey that does NOT report coverage
        within a 1 degree circle around the above-defined source position. Otherwise, returned
        resource lists will be identical.
        """
        # Set Common Args
        vo_plugin = self._init_voplugin_M51(imviz_helper)

        # Retrieve registry options with filtering on
        vo_plugin.resource_filter_coverage = True
        vo_plugin.query_registry_resources()
        assert vo_plugin.resources_loading is False
        filtered_resources = vo_plugin.resource.choices
        assert len(filtered_resources) > 0

        # Retrieve registry options with filtering off
        vo_plugin.resource_filter_coverage = False
        vo_plugin.query_registry_resources()
        assert vo_plugin.resources_loading is False
        nonfiltered_resources = vo_plugin.resource.choices

        # Nonfiltered resources should be more than filtered resources
        assert len(nonfiltered_resources) > len(filtered_resources)

    def test_target_lookup_warnings(self, imviz_helper):
        """
        Tests that appropriate errors and guardrails protect the user
        when a provided source is irresolvable
        """
        expected_error_msg = "Unable to resolve source coordinates"
        vo_plugin = imviz_helper.plugins[vo_plugin_label]._obj

        # Manually set the source to a fake target
        vo_plugin.viewer.selected = "Manual"
        vo_plugin.source = "ThisIsAFakeTargetThatWontResolveToAnything"
        vo_plugin.radius = 1
        vo_plugin.radius_unit.selected = "deg"

        # If we have coverage filtering on, we should get an error
        vo_plugin.resource_filter_coverage = True
        with pytest.raises(LookupError, match=expected_error_msg):
            vo_plugin.waveband.selected = "optical"
            vo_plugin.query_registry_resources()
            assert any(
                expected_error_msg in d["text"]
                for d in imviz_helper.app.state.snackbar_history
            )
            assert len(vo_plugin.resource.choices) == 0

        # By clearing coverage filtering, we should now be able to query the registry
        # and return the full list of available resources:
        vo_plugin.resource_filter_coverage = False
        vo_plugin.query_registry_resources()
        assert len(vo_plugin.resource.choices) > 0

        # However, if we try to query a resource, we should be prevented
        # since the source still isn't resolvable.
        # Clear existing messages
        imviz_helper.app.state.snackbar_history = []
        vo_plugin.resource.selected = "HST.M51"
        with pytest.raises(LookupError, match=expected_error_msg):
            vo_plugin.vue_query_resource()
            assert any(
                expected_error_msg in d["text"]
                for d in imviz_helper.app.state.snackbar_history
            )

    @pytest.mark.filterwarnings("ignore::astropy.wcs.wcs.FITSFixedWarning")
    @pytest.mark.filterwarnings("ignore:Some non-standard WCS keywords were excluded")
    def test_HSTM51_load_data(self, imviz_helper):
        """
        Tests the following:
        * The full plugin by filling out the form and loading a data product into Imviz.
        * The plugin warns user about potentially misaligned data layers
            when loading data and WCS linking is not enabled.
        * User gets properly notified when a load_data error occurs
        """
        # Set Common Args
        vo_plugin = self._init_voplugin_M51(imviz_helper)

        # Select HST.M51 survey
        # Coverage not implemented for HST.M51
        vo_plugin.resource_filter_coverage = False
        vo_plugin.query_registry_resources()
        assert "HST.M51" in vo_plugin.resource.choices
        vo_plugin.resource.selected = "HST.M51"

        # Query resource
        vo_plugin.vue_query_resource()
        assert len(vo_plugin.table.items) > 0

        # Add poisoned fake result
        fake_result = fake_siaresult(
            {
                "title": "Fake Title",
                "instr": "Fake Instrument",
                "dateobs": "Fake Dateobs",
            }
        )
        vo_plugin._populate_table([fake_result])

        # Put the fake result first to make sure we hit the fake result before the real one
        vo_plugin.table.selected_rows = [
            vo_plugin.table.items[-1],
            vo_plugin.table.items[0],
        ]
        assert vo_plugin.table.selected_rows[0]["URL"] == fake_result.getdataurl()

        # Load first data product and fake result
        vo_plugin.load_selected_data()
        assert vo_plugin.data_loading is False
        # Test that user was warned about failed file loading on fake result
        assert any(
            "Unable to load file to viewer" in d["text"]
            for d in imviz_helper.app.state.snackbar_history
        )
        # But also test that didn't prevent us from loading the valid results
        assert len(imviz_helper.app.data_collection) == 1
        assert "M51_HST.M51" in imviz_helper.data_labels[0]

        # Load second data product
        imviz_helper.app.state.snackbar_history = []  # Clear snackbar warnings
        # User should be warned about misaligned data if WCS linking isn't set
        # and there's already data in the data collection
        assert imviz_helper.plugins["Orientation"].align_by == "Pixels"
        vo_plugin.table.selected_rows = [
            vo_plugin.table.items[0]
        ]  # Select first entry
        vo_plugin.load_selected_data()
        assert vo_plugin.data_loading is False
        assert any(
            "WCS linking is not enabled; data layers may not be aligned" in d["text"]
            for d in imviz_helper.app.state.snackbar_history
        )

        # Load third data product
        imviz_helper.app.state.snackbar_history = []  # Clear snackbar warnings
        # If we switch to WCS linking, we shouldn't get a warning anymore
        # since the data will be aligned
        imviz_helper.plugins["Orientation"].align_by = "WCS"
        vo_plugin.table.selected_rows = [
            vo_plugin.table.items[0]
        ]  # Select first entry
        vo_plugin.load_selected_data()
        assert vo_plugin.data_loading is False
        assert all(
            "WCS linking is not enabled; data layers may not be aligned"
            not in d["text"]
            for d in imviz_helper.app.state.snackbar_history
        )



import pytest
from traitlets import HasTraits, TraitError

from jdaviz.core.custom_traitlets import IntHandleEmpty, FloatHandleEmpty


class Foo(HasTraits):
    int_handle_empty = IntHandleEmpty(0)
    int_handle_empty_replace = IntHandleEmpty(0, replace_with_default=True)
    float_handle_empty = FloatHandleEmpty()


def test_inthandleempty():
    foo = Foo()

    foo.int_handle_empty = 1
    foo.int_handle_empty = ''
    assert foo.int_handle_empty == ''
    with pytest.raises(
            TraitError,
            match=r"The 'int_handle_empty' trait of a Foo instance expected an int, not the str 'blah'\."):  # noqa
        foo.int_handle_empty = 'blah'

    foo.int_handle_empty_replace = 1
    foo.int_handle_empty_replace = ''
    assert foo.int_handle_empty_replace == 0

    foo.float_handle_empty = 1.2



from jdaviz.configs.imviz.tests.utils import BaseImviz_WCS_WCS


# This applies to all viz but testing with Imviz should be enough.
class TestImviz_WCS_WCS(BaseImviz_WCS_WCS):
    def test_imviz_zoom_level(self):
        v = self.imviz.viewers['imviz-0']
        assert v._obj.state.x_min == -0.5
        assert v._obj.state.x_max == 9.5

        v.zoom(2)

        assert v._obj.state.x_min == 1.5
        assert v._obj.state.x_max == 6.5

    def test_imviz_viewers(self):
        self.imviz.create_image_viewer()
        self.imviz.create_image_viewer()

        # regression test for https://github.com/spacetelescope/jdaviz/pull/2624
        assert len(self.imviz.viewers) == 3


def test_specviz_zoom_level(specviz_helper):
    v = specviz_helper.viewers['spectrum-viewer']
    v.set_limits(x_min=1, x_max=2, y_min=1, y_max=2)
    assert v._obj.state.x_min == 1
    assert v._obj.state.x_max == 2
    assert v._obj.state.y_min == 1
    assert v._obj.state.y_max == 2


def test_specviz_data_labels(specviz_helper, spectrum1d):
    label = "Test 1D Spectrum"
    specviz_helper.load_data(spectrum1d, data_label=label)

    assert specviz_helper.data_labels == [label]
    assert specviz_helper.viewers['spectrum-viewer'].data_menu.data_labels_loaded == [label]
    assert specviz_helper.viewers['spectrum-viewer'].data_menu.data_labels_visible == [label]


def test_toggle_api_hints(specviz_helper):
    assert specviz_helper.app.state.show_api_hints is False
    specviz_helper.toggle_api_hints()
    assert specviz_helper.app.state.show_api_hints is True
    specviz_helper.toggle_api_hints(True)
    assert specviz_helper.app.state.show_api_hints is True
    specviz_helper.toggle_api_hints()
    assert specviz_helper.app.state.show_api_hints is False



from traitlets import List

from jdaviz.core.events import NewViewerMessage
from jdaviz.core.template_mixin import TemplateMixin
from jdaviz.core.registries import tool_registry
from jdaviz.configs.imviz.plugins.viewers import ImvizImageView

__all__ = ['ImageViewerCreator']


@tool_registry('g-image-viewer-creator')
class ImageViewerCreator(TemplateMixin):

    template_file = __file__, "image_viewer_creator.vue"
    viewer_types = List([]).tag(sync=True)

    def vue_create_image_viewer(self, *args, **kwargs):

        new_viewer_message = NewViewerMessage(
            ImvizImageView, data=None, sender=self)

        self.hub.broadcast(new_viewer_message)



"""This module tests region handling specific to Cubeviz.
Generic handling logic already covered in
jdaviz/configs/imviz/tests/test_regions.py
"""
import numpy as np
import pytest
from astropy import units as u
from astropy.coordinates import SkyCoord
from regions import PixCoord, CirclePixelRegion, CircleSkyRegion, EllipsePixelRegion
from specutils import Spectrum1D, SpectralRegion

from jdaviz.configs.imviz.tests.test_regions import BaseRegionHandler


class TestLoadRegions(BaseRegionHandler):
    @pytest.fixture(autouse=True)
    def setup_class(self, cubeviz_helper, image_cube_hdu_obj_microns):
        self.cubeviz = cubeviz_helper
        cubeviz_helper.load_data(image_cube_hdu_obj_microns, data_label='has_microns')
        self.viewer = cubeviz_helper.default_viewer._obj  # This is used in BaseRegionHandler
        self.spectrum_viewer = cubeviz_helper.app.get_viewer(
            cubeviz_helper._default_spectrum_viewer_reference_name
        )

    def teardown_method(self, method):
        """Clear all the subsets for the next test method."""
        self.cubeviz.app.delete_subsets()

    def test_regions_mask(self):
        mask = np.zeros((9, 10), dtype=np.bool_)
        mask[0, 0] = True
        bad_regions = self.cubeviz.plugins['Subset Tools'].import_region(
            [mask], return_bad_regions=True)

        # TODO: Update expected results if we ever support masked Subset in Cubeviz.
        assert len(bad_regions) == 1 and bad_regions[0][1] == 'Mask creation failed'

    def test_regions_pixel(self):
        # A little out-of-bounds should still overlay the overlapped part.
        my_reg = CirclePixelRegion(center=PixCoord(x=6, y=2), radius=5)
        bad_regions = self.cubeviz.plugins['Subset Tools'].import_region(
            [my_reg], return_bad_regions=True)
        assert len(bad_regions) == 0
        self.verify_region_loaded('Subset 1', count=1)
        assert len(self.cubeviz.plugins['Subset Tools'].get_regions()) == 1

    def test_regions_sky_has_wcs(self):
        sky = SkyCoord(205.4397, 27.0035, unit='deg')
        my_reg_sky_1 = CircleSkyRegion(center=sky, radius=0.0004 * u.deg)
        bad_regions = self.cubeviz.plugins['Subset Tools'].import_region(
            my_reg_sky_1, return_bad_regions=True)

        # TODO: Update expected results when we support sky regions in Cubeviz.
        assert len(bad_regions) == 1 and bad_regions[0][1] == 'Sky region provided but data has no valid WCS'  # noqa

    def test_spatial_spectral_mix(self):
        # Draw ellipse and wavelength range.
        unit = u.Unit(self.cubeviz.plugins['Unit Conversion'].spectral_unit.selected)
        self.cubeviz.plugins['Subset Tools'].import_region([
            EllipsePixelRegion(center=PixCoord(x=4.5, y=4.0), width=9.0, height=8.0),
            SpectralRegion(4.892 * unit, 4.896 * unit)
        ], combination_mode="new")
        self.cubeviz.app.session.edit_subset_mode.edit_subset = None

        # Get spatial regions only.
        st = self.cubeviz.plugins['Subset Tools']._obj
        spatial_subsets_as_regions = st.get_regions(region_type='spatial')
        assert list(spatial_subsets_as_regions.keys()) == ['Subset 1'], spatial_subsets_as_regions
        assert isinstance(spatial_subsets_as_regions['Subset 1'], EllipsePixelRegion)
        # ensure agreement between app.get_subsets and subset_tools.get_regions
        ss = self.cubeviz.app.get_subsets()
        assert ss['Subset 1'][0]['region'] == spatial_subsets_as_regions['Subset 1']

        # NOTE: This does not test that spectrum from Subset is actually scientifically accurate.
        # Get spectral regions only.
        # https://github.com/spacetelescope/jdaviz/issues/1584
        with pytest.warns(UserWarning, match='Applying the value from the redshift slider'):
            spectral_subsets = self.cubeviz.specviz.get_spectra()
        assert list(spectral_subsets.keys()) == ['Spectrum (sum)',
                                                 'Spectrum (sum) (Subset 2)',
                                                 'Spectrum (Subset 1, sum)',
                                                 'Spectrum (Subset 1, sum) (Subset 2)']
        for sp in spectral_subsets.values():
            assert isinstance(sp, Spectrum1D)



import pytest
from jdaviz.configs.cubeviz.plugins.slice.slice import Slice
from jdaviz.configs.imviz.plugins.parsers import HAS_ROMAN_DATAMODELS


@pytest.mark.skipif(not HAS_ROMAN_DATAMODELS, reason="roman_datamodels is not installed")
def test_slice_roman(rampviz_helper, roman_level_1_ramp):
    _slice(rampviz_helper, roman_level_1_ramp)


def test_slice_jwst(rampviz_helper, jwst_level_1b_ramp):
    _slice(rampviz_helper, jwst_level_1b_ramp)


def _slice(helper, ramp_cube):
    app = helper.app
    sl = Slice(app=app)

    # No data yet
    assert len(sl.slice_selection_viewers) == 2  # group-viewer, diff-viewer
    assert len(sl.slice_indicator_viewers) == 1  # integration-viewer
    assert len(sl.valid_indicator_values_sorted) == 0
    assert len(sl.valid_selection_values_sorted) == 0

    # Make sure nothing crashes if plugin used without data]
    sl.vue_play_next()
    sl.vue_play_start_stop()
    assert not sl.is_playing

    helper.load_data(ramp_cube, data_label='test')
    app.add_data_to_viewer("group-viewer", "test[DATA]")
    app.add_data_to_viewer("diff-viewer", "test[DATA]")
    app.add_data_to_viewer("integration-viewer", "Ramp (mean)")
    sv = helper.viewers['integration-viewer']._obj

    # sample ramp only has 10 groups
    assert len(sv.slice_values) == 10
    assert len(sl.valid_indicator_values_sorted) == 10
    slice_values = sl.valid_selection_values_sorted
    assert len(slice_values) == 10

    assert sl.value == slice_values[len(slice_values) // 2]
    assert helper.app.get_viewer("group-viewer").slice == len(slice_values) // 2
    assert helper.app.get_viewer("group-viewer").state.slices[-1] == 5
    assert helper.app.get_viewer("diff-viewer").state.slices[-1] == 5
    helper.select_group(slice_values[0])
    assert helper.app.get_viewer("group-viewer").slice == 0
    assert sl.value == slice_values[0]

    helper.select_group(slice_values[1])
    assert sl.value == slice_values[1]

    # Retrieve updated slice_values
    slice_values = sl.valid_selection_values_sorted

    # Test player buttons API

    sl.vue_goto_first()
    assert sl.value == slice_values[0]

    sl.vue_goto_last()
    assert sl.value == slice_values[-1]

    sl.vue_play_next()  # Should automatically wrap to beginning
    assert sl.value == slice_values[0]

    sl.vue_play_start_stop()  # Start
    assert sl.is_playing
    assert sl._player.is_alive()
    sl.vue_play_next()  # Should be no-op
    sl.vue_goto_last()  # Should be no-op
    sl.vue_goto_first()  # Should be no-op
    sl.vue_play_start_stop()  # Stop
    assert not sl.is_playing
    assert not sl._player
    # NOTE: Hard to check sl.slice here because it is non-deterministic.


@pytest.mark.skipif(not HAS_ROMAN_DATAMODELS, reason="roman_datamodels is not installed")
def test_indicator_settings_roman(rampviz_helper, roman_level_1_ramp):
    _indicator_settings(rampviz_helper, roman_level_1_ramp)


def test_indicator_settings_jwst(rampviz_helper, jwst_level_1b_ramp):
    _indicator_settings(rampviz_helper, jwst_level_1b_ramp)


def _indicator_settings(helper, ramp):
    helper.load_data(ramp, data_label='test')
    app = helper.app
    app.add_data_to_viewer("group-viewer", "test[DATA]")
    app.add_data_to_viewer("integration-viewer", "Ramp (mean)")
    sl = helper.plugins['Slice']._obj
    sv = app.get_viewer('integration-viewer')
    indicator = sv.slice_indicator

    assert sl.show_indicator is True
    assert indicator._show_if_inactive is True
    assert sl.show_value is True
    assert indicator.label.visible is True

    sl.show_indicator = False
    assert indicator._show_if_inactive is False

    sl.show_value = False
    assert indicator.label.visible is False


@pytest.mark.skipif(not HAS_ROMAN_DATAMODELS, reason="roman_datamodels is not installed")
def test_init_slice_roman(rampviz_helper, roman_level_1_ramp):
    _init_slice(rampviz_helper, roman_level_1_ramp)


def test_init_slice_jwst(rampviz_helper, jwst_level_1b_ramp):
    _init_slice(rampviz_helper, jwst_level_1b_ramp)


def _init_slice(helper, ramp):
    helper.load_data(ramp, data_label='test')

    fv = helper.app.get_viewer('group-viewer')
    sl = helper.plugins['Slice']
    slice_values = sl._obj.valid_selection_values_sorted

    assert sl.value == slice_values[len(slice_values)//2]
    assert fv.slice == 5
    assert fv.state.slices == (0, 0, 5)



from traitlets import Bool

from jdaviz.core.registries import tool_registry
from jdaviz.core.template_mixin import TemplateMixin
from jdaviz.core.events import RowLockMessage

__all__ = ['RowLock']


@tool_registry('g-row-lock')
class RowLock(TemplateMixin):
    template_file = __file__, "row_lock.vue"
    is_locked = Bool(False).tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Watch for messages from Specviz helper redshift functions
        self.session.hub.subscribe(self, RowLockMessage,
                                   handler=self._row_lock_changed)

    def _row_lock_changed(self, msg):
        self.is_locked = msg.is_locked

    def vue_toggle_lock(self, event):
        # Send the RowLockMessage which will update is_locked in _row_lock_changed
        # but will also set the application state in the Mosviz helper
        msg = RowLockMessage(not self.is_locked, sender=self)
        self.app.hub.broadcast(msg)
